@@ -1,21 +1,12 @@
-#include <stdlib.h>
-#include <unistd.h>
-#include <string.h>
-#include <sys/socket.h>
-#include <netdb.h>
-#include <sys/types.h>
-#include <sys/socket.h>
-#include <netinet/in.h>
-#include <arpa/inet.h>
-#include <time.h>
 #include "owl.h"
-
-static const char fileIdent[] = "$Id$";
+#include "filterproc.h"
+#include <sys/socket.h>
+#include <arpa/inet.h>
 
 static owl_fmtext_cache fmtext_cache[OWL_FMTEXT_CACHE_SIZE];
 static owl_fmtext_cache * fmtext_cache_next = fmtext_cache;
 
-void owl_message_init_fmtext_cache ()
+void owl_message_init_fmtext_cache(void)
 {
     int i;
     for(i = 0; i < OWL_FMTEXT_CACHE_SIZE; i++) {
@@ -24,7 +15,7 @@ void owl_message_init_fmtext_cache ()
     }
 }
 
-owl_fmtext_cache * owl_message_next_fmtext() /*noproto*/
+static owl_fmtext_cache *owl_message_next_fmtext(void)
 {
     owl_fmtext_cache * f = fmtext_cache_next;
     if(fmtext_cache_next->message != NULL) {
@@ -41,15 +32,18 @@ void owl_message_init(owl_message *m)
   m->id=owl_global_get_nextmsgid(&g);
   owl_message_set_direction_none(m);
   m->delete=0;
-  m->zwriteline=NULL;
+
+#ifdef HAVE_LIBZEPHYR
+  m->has_notice = false;
+#endif
 
   owl_message_set_hostname(m, "");
-  owl_list_create(&(m->attributes));
+  m->attributes = g_ptr_array_new();
   
   /* save the time */
-  m->time=time(NULL);
-  m->timestr=owl_strdup(ctime(&(m->time)));
-  m->timestr[strlen(m->timestr)-1]='\0';
+  m->time = time(NULL);
+  m->timestr = g_strdup(ctime(&m->time));
+  m->timestr[strlen(m->timestr)-1] = '\0';
 
   m->fmtext = NULL;
 }
@@ -57,26 +51,27 @@ void owl_message_init(owl_message *m)
 /* add the named attribute to the message.  If an attribute with the
  * name already exists, replace the old value with the new value
  */
-void owl_message_set_attribute(owl_message *m, char *attrname, char *attrvalue)
+void owl_message_set_attribute(owl_message *m, const char *attrname, const char *attrvalue)
 {
-  int i, j;
+  int i;
   owl_pair *p = NULL, *pair = NULL;
 
+  attrname = g_intern_string(attrname);
+
   /* look for an existing pair with this key, */
-  j=owl_list_get_size(&(m->attributes));
-  for (i=0; i<j; i++) {
-    p=owl_list_get_element(&(m->attributes), i);
-    if (!strcmp(owl_pair_get_key(p), attrname)) {
-      owl_free(owl_pair_get_value(p));
+  for (i = 0; i < m->attributes->len; i++) {
+    p = m->attributes->pdata[i];
+    if (owl_pair_get_key(p) == attrname) {
+      g_free(owl_pair_get_value(p));
       pair = p;
       break;
     }
   }
 
   if(pair ==  NULL) {
-    pair = owl_malloc(sizeof(owl_pair));
-    owl_pair_create(pair, owl_global_intern(&g, attrname), NULL);
-    owl_list_append_element(&(m->attributes), pair);
+    pair = g_slice_new(owl_pair);
+    owl_pair_create(pair, attrname, NULL);
+    g_ptr_array_add(m->attributes, pair);
   }
   owl_pair_set_value(pair, owl_validate_or_convert(attrvalue));
 }
@@ -84,15 +79,21 @@ void owl_message_set_attribute(owl_message *m, char *attrname, char *attrvalue)
 /* return the value associated with the named attribute, or NULL if
  * the attribute does not exist
  */
-char *owl_message_get_attribute_value(owl_message *m, char *attrname)
+const char *owl_message_get_attribute_value(const owl_message *m, const char *attrname)
 {
-  int i, j;
+  int i;
   owl_pair *p;
+  GQuark quark;
 
-  j=owl_list_get_size(&(m->attributes));
-  for (i=0; i<j; i++) {
-    p=owl_list_get_element(&(m->attributes), i);
-    if (!strcmp(owl_pair_get_key(p), attrname)) {
+  quark = g_quark_try_string(attrname);
+  if (quark == 0)
+    /* don't bother inserting into string table */
+    return NULL;
+  attrname = g_quark_to_string(quark);
+
+  for (i = 0; i < m->attributes->len; i++) {
+    p = m->attributes->pdata[i];
+    if (owl_pair_get_key(p) == attrname) {
       return(owl_pair_get_value(p));
     }
   }
@@ -109,24 +110,31 @@ char *owl_message_get_attribute_value(owl_message *m, char *attrname)
  * the 'info' function.  Later there should just be a generic
  * function to indent fmtext.
  */
-void owl_message_attributes_tofmtext(owl_message *m, owl_fmtext *fm) {
-  int i, j;
+void owl_message_attributes_tofmtext(const owl_message *m, owl_fmtext *fm) {
+  int i;
   owl_pair *p;
-  char *buff;
+  char *buff, *tmpbuff;
 
   owl_fmtext_init_null(fm);
 
-  j=owl_list_get_size(&(m->attributes));
-  for (i=0; i<j; i++) {
-    p=owl_list_get_element(&(m->attributes), i);
-    buff=owl_sprintf("  %-15.15s: %-35.35s\n", owl_pair_get_key(p), owl_pair_get_value(p));
+  for (i = 0; i < m->attributes->len; i++) {
+    p = m->attributes->pdata[i];
+
+    buff = g_strdup(owl_pair_get_value(p));
+    if (buff) {
+      tmpbuff = owl_text_indent(buff, 19, false);
+      g_free(buff);
+      buff = g_strdup_printf("  %-15.15s: %s\n", owl_pair_get_key(p), tmpbuff);
+      g_free(tmpbuff);
+    }
+
     if(buff == NULL) {
-      buff=owl_sprintf("  %-15.15s: %-35.35s\n", owl_pair_get_key(p), "<error>");
+      buff = g_strdup_printf("  %-15.15s: %s\n", owl_pair_get_key(p), "<error>");
       if(buff == NULL)
-        buff=owl_strdup("   <error>\n");
+        buff=g_strdup("   <error>\n");
     }
     owl_fmtext_append_normal(fm, buff);
-    owl_free(buff);
+    g_free(buff);
   }
 }
 
@@ -147,8 +155,8 @@ owl_fmtext *owl_message_get_fmtext(owl_message *m)
 
 void owl_message_format(owl_message *m)
 {
-  owl_style *s;
-  owl_view *v;
+  const owl_style *s;
+  const owl_view *v;
 
   if (!m->fmtext) {
     m->fmtext = owl_message_next_fmtext();
@@ -161,100 +169,100 @@ void owl_message_format(owl_message *m)
   }
 }
 
-void owl_message_set_class(owl_message *m, char *class)
+void owl_message_set_class(owl_message *m, const char *class)
 {
   owl_message_set_attribute(m, "class", class);
 }
 
-char *owl_message_get_class(owl_message *m)
+const char *owl_message_get_class(const owl_message *m)
 {
-  char *class;
+  const char *class;
 
   class=owl_message_get_attribute_value(m, "class");
   if (!class) return("");
   return(class);
 }
 
-void owl_message_set_instance(owl_message *m, char *inst)
+void owl_message_set_instance(owl_message *m, const char *inst)
 {
   owl_message_set_attribute(m, "instance", inst);
 }
 
-char *owl_message_get_instance(owl_message *m)
+const char *owl_message_get_instance(const owl_message *m)
 {
-  char *instance;
+  const char *instance;
 
   instance=owl_message_get_attribute_value(m, "instance");
   if (!instance) return("");
   return(instance);
 }
 
-void owl_message_set_sender(owl_message *m, char *sender)
+void owl_message_set_sender(owl_message *m, const char *sender)
 {
   owl_message_set_attribute(m, "sender", sender);
 }
 
-char *owl_message_get_sender(owl_message *m)
+const char *owl_message_get_sender(const owl_message *m)
 {
-  char *sender;
+  const char *sender;
 
   sender=owl_message_get_attribute_value(m, "sender");
   if (!sender) return("");
   return(sender);
 }
 
-void owl_message_set_zsig(owl_message *m, char *zsig)
+void owl_message_set_zsig(owl_message *m, const char *zsig)
 {
   owl_message_set_attribute(m, "zsig", zsig);
 }
 
-char *owl_message_get_zsig(owl_message *m)
+const char *owl_message_get_zsig(const owl_message *m)
 {
-  char *zsig;
+  const char *zsig;
 
   zsig=owl_message_get_attribute_value(m, "zsig");
   if (!zsig) return("");
   return(zsig);
 }
 
-void owl_message_set_recipient(owl_message *m, char *recip)
+void owl_message_set_recipient(owl_message *m, const char *recip)
 {
   owl_message_set_attribute(m, "recipient", recip);
 }
 
-char *owl_message_get_recipient(owl_message *m)
+const char *owl_message_get_recipient(const owl_message *m)
 {
   /* this is stupid for outgoing messages, we need to fix it. */
 
-  char *recip;
+  const char *recip;
 
   recip=owl_message_get_attribute_value(m, "recipient");
   if (!recip) return("");
   return(recip);
 }
 
-void owl_message_set_realm(owl_message *m, char *realm)
+void owl_message_set_realm(owl_message *m, const char *realm)
 {
   owl_message_set_attribute(m, "realm", realm);
 }
 
-char *owl_message_get_realm(owl_message *m)
+const char *owl_message_get_realm(const owl_message *m)
 {
-  char *realm;
+  const char *realm;
   
   realm=owl_message_get_attribute_value(m, "realm");
   if (!realm) return("");
   return(realm);
 }
 
-void owl_message_set_body(owl_message *m, char *body)
+void owl_message_set_body(owl_message *m, const char *body)
 {
   owl_message_set_attribute(m, "body", body);
 }
 
-char *owl_message_get_body(owl_message *m)
+const char *owl_message_get_body(const owl_message *m)
 {
-  char *body;
+  const char *body;
 
   body=owl_message_get_attribute_value(m, "body");
   if (!body) return("");
@@ -262,14 +270,14 @@ char *owl_message_get_body(owl_message *m)
 }
 
 
-void owl_message_set_opcode(owl_message *m, char *opcode)
+void owl_message_set_opcode(owl_message *m, const char *opcode)
 {
   owl_message_set_attribute(m, "opcode", opcode);
 }
 
-char *owl_message_get_opcode(owl_message *m)
+const char *owl_message_get_opcode(const owl_message *m)
 {
-  char *opcode;
+  const char *opcode;
 
   opcode=owl_message_get_attribute_value(m, "opcode");
   if (!opcode) return("");
@@ -288,18 +296,18 @@ void owl_message_set_islogout(owl_message *m)
   owl_message_set_attribute(m, "loginout", "logout");
 }
 
-int owl_message_is_loginout(owl_message *m)
+int owl_message_is_loginout(const owl_message *m)
 {
-  char *res;
+  const char *res;
 
   res=owl_message_get_attribute_value(m, "loginout");
   if (!res) return(0);
   return(1);
 }
 
-int owl_message_is_login(owl_message *m)
+int owl_message_is_login(const owl_message *m)
 {
-  char *res;
+  const char *res;
 
   res=owl_message_get_attribute_value(m, "loginout");
   if (!res) return(0);
@@ -308,9 +316,9 @@ int owl_message_is_login(owl_message *m)
 }
 
 
-int owl_message_is_logout(owl_message *m)
+int owl_message_is_logout(const owl_message *m)
 {
-  char *res;
+  const char *res;
 
   res=owl_message_get_attribute_value(m, "loginout");
   if (!res) return(0);
@@ -323,31 +331,24 @@ void owl_message_set_isprivate(owl_message *m)
   owl_message_set_attribute(m, "isprivate", "true");
 }
 
-int owl_message_is_private(owl_message *m)
+int owl_message_is_private(const owl_message *m)
 {
-  char *res;
+  const char *res;
 
   res=owl_message_get_attribute_value(m, "isprivate");
   if (!res) return(0);
   return !strcmp(res, "true");
 }
 
-char *owl_message_get_timestr(owl_message *m)
+const char *owl_message_get_timestr(const owl_message *m)
 {
   if (m->timestr) return(m->timestr);
   return("");
 }
 
-/* caller must free the return */
-char *owl_message_get_shorttimestr(owl_message *m)
+CALLER_OWN char *owl_message_format_time(const owl_message *m)
 {
-  struct tm *tmstruct;
-  char *out;
-
-  tmstruct=localtime(&(m->time));
-  out=owl_sprintf("%2.2i:%2.2i", tmstruct->tm_hour, tmstruct->tm_min);
-  if (out) return(out);
-  return("??:??");
+  return owl_util_format_time(localtime(&m->time));
 }
 
 void owl_message_set_type_admin(owl_message *m)
@@ -370,56 +371,50 @@ void owl_message_set_type_aim(owl_message *m)
   owl_message_set_attribute(m, "type", "AIM");
 }
 
-void owl_message_set_type(owl_message *m, char* type)
+void owl_message_set_type(owl_message *m, const char* type)
 {
   owl_message_set_attribute(m, "type", type);
 }
 
-int owl_message_is_type(owl_message *m, char *type) {
-  char * t = owl_message_get_attribute_value(m, "type");
+int owl_message_is_type(const owl_message *m, const char *type) {
+  const char * t = owl_message_get_attribute_value(m, "type");
   if(!t) return 0;
   return !strcasecmp(t, type);
 }
 						
-int owl_message_is_type_admin(owl_message *m)
+int owl_message_is_type_admin(const owl_message *m)
 {
   return owl_message_is_type(m, "admin");
 }
 
-int owl_message_is_type_generic(owl_message *m)
-{
-  char * t = owl_message_get_attribute_value(m, "type");
-  return (t == NULL);
-}
-
-int owl_message_is_type_zephyr(owl_message *m)
+int owl_message_is_type_zephyr(const owl_message *m)
 {
   return owl_message_is_type(m, "zephyr");
 }
 
-int owl_message_is_type_aim(owl_message *m)
+int owl_message_is_type_aim(const owl_message *m)
 {
   return owl_message_is_type(m, "aim");
 }
 
 /* XXX TODO: deprecate this */
-int owl_message_is_type_jabber(owl_message *m)
+int owl_message_is_type_jabber(const owl_message *m)
 {
   return owl_message_is_type(m, "jabber");
 }
 
-int owl_message_is_type_loopback(owl_message *m)
+int owl_message_is_type_loopback(const owl_message *m)
 {
   return owl_message_is_type(m, "loopback");
 }
 
-int owl_message_is_pseudo(owl_message *m)
+int owl_message_is_pseudo(const owl_message *m)
 {
   if (owl_message_get_attribute_value(m, "pseudo")) return(1);
   return(0);
 }
 
-char *owl_message_get_text(owl_message *m)
+const char *owl_message_get_text(owl_message *m)
 {
   owl_message_format(m);
   return(owl_fmtext_get_text(&(m->fmtext->fmtext)));
@@ -445,19 +440,19 @@ void owl_message_set_direction(owl_message *m, int direction)
   m->direction=direction;
 }
 
-int owl_message_is_direction_in(owl_message *m)
+int owl_message_is_direction_in(const owl_message *m)
 {
   if (m->direction==OWL_MESSAGE_DIRECTION_IN) return(1);
   return(0);
 }
 
-int owl_message_is_direction_out(owl_message *m)
+int owl_message_is_direction_out(const owl_message *m)
 {
   if (m->direction==OWL_MESSAGE_DIRECTION_OUT) return(1);
   return(0);
 }
 
-int owl_message_is_direction_none(owl_message *m)
+int owl_message_is_direction_none(const owl_message *m)
 {
   if (m->direction==OWL_MESSAGE_DIRECTION_NONE) return(1);
   return(0);
@@ -482,20 +477,19 @@ void owl_message_unmark_delete(owl_message *m)
   m->delete=0;
 }
 
-char *owl_message_get_zwriteline(owl_message *m)
+const char *owl_message_get_zwriteline(const owl_message *m)
 {
-  if(!m->zwriteline)
-    return "";
-  return(m->zwriteline);
+  const char *z = owl_message_get_attribute_value(m, "zwriteline");
+  if (!z) return "";
+  return z;
 }
 
-void owl_message_set_zwriteline(owl_message *m, char *line)
+void owl_message_set_zwriteline(owl_message *m, const char *line)
 {
-  if(m->zwriteline) owl_free(m->zwriteline);
-  m->zwriteline=owl_strdup(line);
+  owl_message_set_attribute(m, "zwriteline", line);
 }
 
-int owl_message_is_delete(owl_message *m)
+int owl_message_is_delete(const owl_message *m)
 {
   if (m == NULL) return(0);
   if (m->delete==1) return(1);
@@ -503,23 +497,23 @@ int owl_message_is_delete(owl_message *m)
 }
 
 #ifdef HAVE_LIBZEPHYR
-ZNotice_t *owl_message_get_notice(owl_message *m)
+const ZNotice_t *owl_message_get_notice(const owl_message *m)
 {
-  return(&(m->notice));
+  return m->has_notice ? &m->notice : NULL;
 }
 #else
-void *owl_message_get_notice(owl_message *m)
+void *owl_message_get_notice(const owl_message *m)
 {
   return(NULL);
 }
 #endif
 
-void owl_message_set_hostname(owl_message *m, char *hostname)
+void owl_message_set_hostname(owl_message *m, const char *hostname)
 {
-  m->hostname=owl_global_intern(&g, hostname);
+  m->hostname = g_intern_string(hostname);
 }
 
-char *owl_message_get_hostname(owl_message *m)
+const char *owl_message_get_hostname(const owl_message *m)
 {
   return(m->hostname);
 }
@@ -536,18 +530,16 @@ void owl_message_curs_waddstr(owl_message *m, WINDOW *win, int aline, int bline,
   
   owl_fmtext_truncate_lines(&(m->fmtext->fmtext), aline, bline-aline, &a);
   owl_fmtext_truncate_cols(&a, acol, bcol, &b);
-  owl_fmtext_colorize(&b, fgcolor);
-  owl_fmtext_colorizebg(&b, bgcolor);
 
-  owl_fmtext_curs_waddstr(&b, win);
+  owl_fmtext_curs_waddstr(&b, win, OWL_FMTEXT_ATTR_NONE, fgcolor, bgcolor);
 
-  owl_fmtext_free(&a);
-  owl_fmtext_free(&b);
+  owl_fmtext_cleanup(&a);
+  owl_fmtext_cleanup(&b);
 }
 
-int owl_message_is_personal(owl_message *m)
+int owl_message_is_personal(const owl_message *m)
 {
-  owl_filter * f = owl_global_get_filter(&g, "personal");
+  const owl_filter * f = owl_global_get_filter(&g, "personal");
   if(!f) {
       owl_function_error("personal filter is not defined");
       return (0);
@@ -555,15 +547,15 @@ int owl_message_is_personal(owl_message *m)
   return owl_filter_message_match(f, m);
 }
 
-int owl_message_is_question(owl_message *m)
+int owl_message_is_question(const owl_message *m)
 {
   if(!owl_message_is_type_admin(m)) return 0;
   if(owl_message_get_attribute_value(m, "question") != NULL) return 1;
   return 0;
 }
 
-int owl_message_is_answered(owl_message *m) {
-  char *q;
+int owl_message_is_answered(const owl_message *m) {
+  const char *q;
   if(!owl_message_is_question(m)) return 0;
   q = owl_message_get_attribute_value(m, "question");
   if(!q) return 0;
@@ -574,27 +566,7 @@ void owl_message_set_isanswered(owl_message *m) {
   owl_message_set_attribute(m, "question", "answered");
 }
 
-int owl_message_is_from_me(owl_message *m)
-{
-  if (owl_message_is_type_zephyr(m)) {
-    if (!strcasecmp(owl_message_get_sender(m), owl_zephyr_get_sender())) {
-      return(1);
-    } else {
-      return(0);
-    }
-  } else if (owl_message_is_type_aim(m)) {
-    if (!strcasecmp(owl_message_get_sender(m), owl_global_get_aim_screenname(&g))) {
-      return(1);
-    } else {
-      return(0);
-    }
-  } else if (owl_message_is_type_admin(m)) {
-    return(0);
-  }
-  return(0);
-}
-
-int owl_message_is_mail(owl_message *m)
+int owl_message_is_mail(const owl_message *m)
 {
   if (owl_message_is_type_zephyr(m)) {
     if (!strcasecmp(owl_message_get_class(m), "mail") && owl_message_is_private(m)) {
@@ -606,109 +578,66 @@ int owl_message_is_mail(owl_message *m)
   return(0);
 }
 
-int owl_message_is_ping(owl_message *m)
-{
-  if (owl_message_is_type_zephyr(m)) {
-    if (!strcasecmp(owl_message_get_opcode(m), "ping")) {
-      return(1);
-    } else {
-      return(0);
-    }
-  }
-  return(0);
-}
-
-int owl_message_is_burningears(owl_message *m)
-{
-  /* we should add a global to cache the short zsender */
-  char sender[LINE], *ptr;
-
-  /* if the message is from us or to us, it doesn't count */
-  if (owl_message_is_from_me(m) || owl_message_is_private(m)) return(0);
-
-  if (owl_message_is_type_zephyr(m)) {
-    strcpy(sender, owl_zephyr_get_sender());
-    ptr=strchr(sender, '@');
-    if (ptr) *ptr='\0';
-  } else if (owl_message_is_type_aim(m)) {
-    strcpy(sender, owl_global_get_aim_screenname(&g));
-  } else {
-    return(0);
-  }
-
-  if (stristr(owl_message_get_body(m), sender)) {
-    return(1);
-  }
-  return(0);
-}
-
 /* caller must free return value. */
-char *owl_message_get_cc(owl_message *m)
+CALLER_OWN char *owl_message_get_cc(const owl_message *m)
 {
-  char *cur, *out, *end;
+  const char *cur;
+  char *out, *end;
 
   cur = owl_message_get_body(m);
   while (*cur && *cur==' ') cur++;
   if (strncasecmp(cur, "cc:", 3)) return(NULL);
   cur+=3;
   while (*cur && *cur==' ') cur++;
-  out = owl_strdup(cur);
+  out = g_strdup(cur);
   end = strchr(out, '\n');
   if (end) end[0] = '\0';
   return(out);
 }
 
 /* caller must free return value */
-char *owl_message_get_cc_without_recipient(owl_message *m)
+CALLER_OWN GList *owl_message_get_cc_without_recipient(const owl_message *m)
 {
-  char *cc, *out, *end, *user, *shortuser, *recip;
+  char *cc, *shortuser, *recip;
+  const char *user;
+  GList *out = NULL;
 
   cc = owl_message_get_cc(m);
   if (cc == NULL)
     return NULL;
 
   recip = short_zuser(owl_message_get_recipient(m));
-  out = owl_malloc(strlen(cc));
-  end = out;
 
   user = strtok(cc, " ");
   while (user != NULL) {
     shortuser = short_zuser(user);
     if (strcasecmp(shortuser, recip) != 0) {
-      strcpy(end, user);
-      end[strlen(user)] = ' ';
-      end += strlen(user) + 1;
+      out = g_list_prepend(out, g_strdup(user));
     }
-    free(shortuser);
+    g_free(shortuser);
     user = strtok(NULL, " ");
   }
-  end[0] = '\0';
 
-  owl_free(recip);
-  owl_free(cc);
-
-  if (strlen(out) == 0) {
-    owl_free(out);
-    out = NULL;
-  }
+  g_free(recip);
+  g_free(cc);
 
   return(out);
 }
 
-int owl_message_get_id(owl_message *m)
+int owl_message_get_id(const owl_message *m)
 {
   return(m->id);
 }
 
-char *owl_message_get_type(owl_message *m) {
-  char * type = owl_message_get_attribute_value(m, "type");
+const char *owl_message_get_type(const owl_message *m) {
+  const char * type = owl_message_get_attribute_value(m, "type");
   if(!type) {
     return "generic";
   }
   return type;
 }
 
-char *owl_message_get_direction(owl_message *m) {
+const char *owl_message_get_direction(const owl_message *m) {
   switch (m->direction) {
   case OWL_MESSAGE_DIRECTION_IN:
     return("in");
@@ -721,7 +650,7 @@ char *owl_message_get_direction(owl_message *m) {
   }
 }
 
-int owl_message_parse_direction(char *d) {
+int owl_message_parse_direction(const char *d) {
   if(!strcmp(d, "in")) {
     return OWL_MESSAGE_DIRECTION_IN;
   } else if(!strcmp(d, "out")) {
@@ -732,7 +661,7 @@ int owl_message_parse_direction(char *d) {
 }
 
 
-char *owl_message_get_login(owl_message *m) {
+const char *owl_message_get_login(const owl_message *m) {
   if (owl_message_is_login(m)) {
     return "login";
   } else if (owl_message_is_logout(m)) {
@@ -743,19 +672,19 @@ char *owl_message_get_login(owl_message *m) {
 }
 
 
-char *owl_message_get_header(owl_message *m) {
+const char *owl_message_get_header(const owl_message *m) {
   return owl_message_get_attribute_value(m, "adminheader");
 }
 
 /* return 1 if the message contains "string", 0 otherwise.  This is
  * case insensitive because the functions it uses are
  */
-int owl_message_search(owl_message *m, char *string)
+int owl_message_search(owl_message *m, const owl_regex *re)
 {
 
   owl_message_format(m); /* is this necessary? */
   
-  return (owl_fmtext_search(&(m->fmtext->fmtext), string));
+  return owl_fmtext_search(&(m->fmtext->fmtext), re, 0) >= 0;
 }
 
 
@@ -763,7 +692,7 @@ int owl_message_search(owl_message *m, char *string)
  *                 0 it's not a login/logout message
  *                 1 it's a login message
  */
-void owl_message_create_aim(owl_message *m, char *sender, char *recipient, char *text, int direction, int loginout)
+void owl_message_create_aim(owl_message *m, const char *sender, const char *recipient, const char *text, int direction, int loginout)
 {
   owl_message_init(m);
   owl_message_set_body(m, text);
@@ -789,7 +718,7 @@ void owl_message_create_aim(owl_message *m, char *sender, char *recipient, char
   }
 }
 
-void owl_message_create_admin(owl_message *m, char *header, char *text)
+void owl_message_create_admin(owl_message *m, const char *header, const char *text)
 {
   owl_message_init(m);
   owl_message_set_type_admin(m);
@@ -798,7 +727,7 @@ void owl_message_create_admin(owl_message *m, char *header, char *text)
 }
 
 /* caller should set the direction */
-void owl_message_create_loopback(owl_message *m, char *text)
+void owl_message_create_loopback(owl_message *m, const char *text)
 {
   owl_message_init(m);
   owl_message_set_type_loopback(m);
@@ -808,11 +737,50 @@ void owl_message_create_loopback(owl_message *m, char *text)
   owl_message_set_isprivate(m);
 }
 
+void owl_message_save_ccs(owl_message *m) {
+  GList *cc;
+  char *tmp;
+
+  cc = owl_message_get_cc_without_recipient(m);
+
+  if (cc != NULL) {
+    GString *recips = g_string_new("");
+    cc = g_list_prepend(cc, short_zuser(owl_message_get_sender(m)));
+    cc = g_list_prepend(cc, short_zuser(owl_message_get_recipient(m)));
+    cc = g_list_sort(cc, (GCompareFunc)strcasecmp);
+
+    while(cc != NULL) {
+      /* Collapse any identical entries */
+      while (cc->next && strcasecmp(cc->data, cc->next->data) == 0) {
+        g_free(cc->data);
+        cc = g_list_delete_link(cc, cc);
+      }
+
+      tmp = short_zuser(cc->data);
+      g_string_append(recips, tmp);
+
+      g_free(tmp);
+      g_free(cc->data);
+      cc = g_list_delete_link(cc, cc);
+
+      if (cc)
+        g_string_append_c(recips, ' ');
+    }
+
+    owl_message_set_attribute(m, "zephyr_ccs", recips->str);
+    g_string_free(recips, true);
+  }
+}
+
 #ifdef HAVE_LIBZEPHYR
-void owl_message_create_from_znotice(owl_message *m, ZNotice_t *n)
+void owl_message_create_from_znotice(owl_message *m, const ZNotice_t *n)
 {
+#ifdef ZNOTICE_SOCKADDR
+  char hbuf[NI_MAXHOST];
+#else /* !ZNOTICE_SOCKADDR */
   struct hostent *hent;
-  char *ptr, *tmp, *tmp2;
+#endif /* ZNOTICE_SOCKADDR */
+  char *tmp, *tmp2;
   int len;
 
   owl_message_init(m);
@@ -821,16 +789,17 @@ void owl_message_create_from_znotice(owl_message *m, ZNotice_t *n)
   owl_message_set_direction_in(m);
   
   /* first save the full notice */
-  memcpy(&(m->notice), n, sizeof(ZNotice_t));
+  m->notice = *n;
+  m->has_notice = true;
 
   /* a little gross, we'll replace \r's with ' ' for now */
   owl_zephyr_hackaway_cr(&(m->notice));
   
   /* save the time, we need to nuke the string saved by message_init */
-  if (m->timestr) owl_free(m->timestr);
-  m->time=n->z_time.tv_sec;
-  m->timestr=owl_strdup(ctime(&(m->time)));
-  m->timestr[strlen(m->timestr)-1]='\0';
+  if (m->timestr) g_free(m->timestr);
+  m->time = n->z_time.tv_sec;
+  m->timestr = g_strdup(ctime(&m->time));
+  m->timestr[strlen(m->timestr)-1] = '\0';
 
   /* set other info */
   owl_message_set_sender(m, n->z_sender);
@@ -844,22 +813,18 @@ void owl_message_create_from_znotice(owl_message *m, ZNotice_t *n)
   }
   owl_message_set_zsig(m, owl_zephyr_get_zsig(n, &len));
 
-  if ((ptr=strchr(n->z_recipient, '@'))!=NULL) {
-    owl_message_set_realm(m, ptr+1);
-  } else {
-    owl_message_set_realm(m, owl_zephyr_get_realm());
-  }
+  owl_message_set_realm(m, zuser_realm(n->z_recipient));
 
   /* Set the "isloginout" attribute if it's a login message */
   if (!strcasecmp(n->z_class, "login") || !strcasecmp(n->z_class, OWL_WEBZEPHYR_CLASS)) {
     if (!strcasecmp(n->z_opcode, "user_login") || !strcasecmp(n->z_opcode, "user_logout")) {
       tmp=owl_zephyr_get_field(n, 1);
       owl_message_set_attribute(m, "loginhost", tmp);
-      owl_free(tmp);
+      g_free(tmp);
 
       tmp=owl_zephyr_get_field(n, 3);
       owl_message_set_attribute(m, "logintty", tmp);
-      owl_free(tmp);
+      g_free(tmp);
     }
 
     if (!strcasecmp(n->z_opcode, "user_login")) {
@@ -870,10 +835,8 @@ void owl_message_create_from_znotice(owl_message *m, ZNotice_t *n)
   }
 
   
-  /* set the "isprivate" attribute if it's a private zephyr.
-   ``private'' means recipient is non-empty and doesn't start wit 
-   `@' */
-  if (*n->z_recipient && *n->z_recipient != '@') {
+  /* set the "isprivate" attribute if it's a private zephyr. */
+  if (owl_zwrite_recip_is_personal(n->z_recipient)) {
     owl_message_set_isprivate(m);
   }
 
@@ -883,54 +846,77 @@ void owl_message_create_from_znotice(owl_message *m, ZNotice_t *n)
     owl_message_set_attribute(m, "isauto", "");
   }
 
-  m->zwriteline=owl_strdup("");
-
   /* save the hostname */
+#ifdef ZNOTICE_SOCKADDR
+  owl_function_debugmsg("About to do getnameinfo");
+  if (getnameinfo(&n->z_sender_sockaddr.sa, sizeof(n->z_sender_sockaddr), hbuf, sizeof(hbuf), NULL, 0, 0) == 0)
+    owl_message_set_hostname(m, hbuf);
+#else /* !ZNOTICE_SOCKADDR */
   owl_function_debugmsg("About to do gethostbyaddr");
-  hent=gethostbyaddr((char *) &(n->z_uid.zuid_addr), sizeof(n->z_uid.zuid_addr), AF_INET);
-  if (hent && hent->h_name) {
+  hent = gethostbyaddr(&n->z_uid.zuid_addr, sizeof(n->z_uid.zuid_addr), AF_INET);
+  if (hent && hent->h_name)
     owl_message_set_hostname(m, hent->h_name);
-  } else {
+  else
     owl_message_set_hostname(m, inet_ntoa(n->z_sender_addr));
-  }
+#endif /* ZNOTICE_SOCKADDR */
 
   /* set the body */
   tmp=owl_zephyr_get_message(n, m);
   if (owl_global_is_newlinestrip(&g)) {
     tmp2=owl_util_stripnewlines(tmp);
     owl_message_set_body(m, tmp2);
-    owl_free(tmp2);
+    g_free(tmp2);
   } else {
     owl_message_set_body(m, tmp);
   }
-  owl_free(tmp);
+  g_free(tmp);
 
-#ifdef OWL_ENABLE_ZCRYPT
   /* if zcrypt is enabled try to decrypt the message */
   if (owl_global_is_zcrypt(&g) && !strcasecmp(n->z_opcode, "crypt")) {
-    char *out;
-    int ret;
+    const char *argv[] = {
+      NULL,
+      "-D",
+      "-c", owl_message_get_class(m),
+      "-i", owl_message_get_instance(m),
+      NULL
+    };
+    char *out = NULL;
+    int rv;
+    int status;
+    char *zcrypt;
 
-    out=owl_malloc(strlen(owl_message_get_body(m))*16+20);
-    ret=owl_zcrypt_decrypt(out, owl_message_get_body(m), owl_message_get_class(m), owl_message_get_instance(m));
-    if (ret==0) {
+    zcrypt = g_build_filename(owl_get_bindir(), "zcrypt", NULL);
+    argv[0] = zcrypt;
+
+    rv = call_filter(argv, owl_message_get_body(m), &out, &status);
+    g_free(zcrypt);
+
+    if(!rv && !status) {
+      int len = strlen(out);
+      if(len >= 8 && !strcmp(out + len - 8, "**END**\n")) {
+        out[len - 8] = 0;
+      }
       owl_message_set_body(m, out);
     } else {
-      owl_free(out);
+      /* Replace the opcode. Otherwise the UI and other bits of code think the
+       * message was encrypted. */
+      owl_message_set_opcode(m, "failed-decrypt");
     }
+    g_free(out);
   }
-#endif  
+
+  owl_message_save_ccs(m);
 }
 #else
-void owl_message_create_from_znotice(owl_message *m, void *n)
+void owl_message_create_from_znotice(owl_message *m, const void *n)
 {
 }
 #endif
 
 /* If 'direction' is '0' it is a login message, '1' is a logout message. */
-void owl_message_create_pseudo_zlogin(owl_message *m, int direction, char *user, char *host, char *time, char *tty)
+void owl_message_create_pseudo_zlogin(owl_message *m, int direction, const char *user, const char *host, const char *time, const char *tty)
 {
-  char *longuser, *ptr;
+  char *longuser;
 
 #ifdef HAVE_LIBZEPHYR
   memset(&(m->notice), 0, sizeof(ZNotice_t));
@@ -959,97 +945,89 @@ void owl_message_create_pseudo_zlogin(owl_message *m, int direction, char *user,
     owl_message_set_islogout(m);
   }
 
-  if ((ptr=strchr(longuser, '@'))!=NULL) {
-    owl_message_set_realm(m, ptr+1);
-  } else {
-    owl_message_set_realm(m, owl_zephyr_get_realm());
-  }
-
-  m->zwriteline=owl_strdup("");
+  owl_message_set_realm(m, zuser_realm(longuser));
 
   owl_message_set_body(m, "<uninitialized>");
 
   /* save the hostname */
   owl_function_debugmsg("create_pseudo_login: host is %s", host ? host : "");
   owl_message_set_hostname(m, host ? host : "");
-  owl_free(longuser);
+  g_free(longuser);
 }
 
-void owl_message_create_from_zwriteline(owl_message *m, char *line, char *body, char *zsig)
+void owl_message_create_from_zwrite(owl_message *m, const owl_zwrite *z, const char *body, int recip_index)
 {
-  owl_zwrite z;
-  int ret;
-  char hostbuff[5000];
+  char *replyline;
   
   owl_message_init(m);
 
-  /* create a zwrite for the purpose of filling in other message fields */
-  owl_zwrite_create_from_line(&z, line);
-
   /* set things */
   owl_message_set_direction_out(m);
   owl_message_set_type_zephyr(m);
   owl_message_set_sender(m, owl_zephyr_get_sender());
-  owl_message_set_class(m, owl_zwrite_get_class(&z));
-  owl_message_set_instance(m, owl_zwrite_get_instance(&z));
-  if (owl_zwrite_get_numrecips(&z)>0) {
-    owl_message_set_recipient(m,
-			      long_zuser(owl_zwrite_get_recip_n(&z, 0))); /* only gets the first user, must fix */
+  owl_message_set_class(m, owl_zwrite_get_class(z));
+  owl_message_set_instance(m, owl_zwrite_get_instance(z));
+  if (recip_index < owl_zwrite_get_numrecips(z)) {
+    char *zuser = owl_zwrite_get_recip_n_with_realm(z, recip_index);
+    char *longzuser = long_zuser(zuser);
+    owl_message_set_recipient(m, longzuser);
+    owl_message_set_realm(m, zuser_realm(longzuser));
+    g_free(longzuser);
+    g_free(zuser);
+  } else {
+    /* TODO: We should maybe munge this into the case above, as that comparison
+     * is a little overly clever. It's also not clear this codepath ever runs
+     * anyway. */
+    const char *realm = owl_zwrite_get_realm(z);
+    owl_message_set_realm(m, realm[0] ? realm : owl_zephyr_get_realm());
   }
-  owl_message_set_opcode(m, owl_zwrite_get_opcode(&z));
-  owl_message_set_realm(m, owl_zwrite_get_realm(&z)); /* also a hack, but not here */
-  m->zwriteline=owl_strdup(line);
+  owl_message_set_opcode(m, owl_zwrite_get_opcode(z));
+
+  /* Although not strictly the zwriteline, anyone using the unsantized version
+   * of it probably has a bug. */
+  replyline = owl_zwrite_get_replyline(z, recip_index);
+  owl_message_set_zwriteline(m, replyline);
+  g_free(replyline);
+
   owl_message_set_body(m, body);
-  owl_message_set_zsig(m, zsig);
+  owl_message_set_zsig(m, owl_zwrite_get_zsig(z));
   
   /* save the hostname */
-  ret=gethostname(hostbuff, MAXHOSTNAMELEN);
-  hostbuff[MAXHOSTNAMELEN]='\0';
-  if (ret) {
-    owl_message_set_hostname(m, "localhost");
-  } else {
-    owl_message_set_hostname(m, hostbuff);
-  }
+  owl_message_set_hostname(m, g_get_host_name());
 
   /* set the "isprivate" attribute if it's a private zephyr. */
-  if (owl_zwrite_is_personal(&z)) {
+  if (owl_zwrite_is_personal(z)) {
     owl_message_set_isprivate(m);
   }
 
-  owl_zwrite_free(&z);
+  owl_message_save_ccs(m);
 }
 
-void owl_message_pretty_zsig(owl_message *m, char *buff)
+void owl_message_cleanup(owl_message *m)
 {
-  /* stick a one line version of the zsig in buff */
-  char *ptr;
-
-  strcpy(buff, owl_message_get_zsig(m));
-  ptr=strchr(buff, '\n');
-  if (ptr) ptr[0]='\0';
-}
-
-void owl_message_free(owl_message *m)
-{
-  int i, j;
+  int i;
   owl_pair *p;
 #ifdef HAVE_LIBZEPHYR    
-  if (owl_message_is_type_zephyr(m) && owl_message_is_direction_in(m)) {
+  if (m->has_notice) {
     ZFreeNotice(&(m->notice));
   }
 #endif
-  if (m->timestr) owl_free(m->timestr);
-  if (m->zwriteline) owl_free(m->zwriteline);
+  if (m->timestr) g_free(m->timestr);
 
   /* free all the attributes */
-  j=owl_list_get_size(&(m->attributes));
-  for (i=0; i<j; i++) {
-    p=owl_list_get_element(&(m->attributes), i);
-    owl_free(owl_pair_get_value(p));
-    owl_free(p);
+  for (i = 0; i < m->attributes->len; i++) {
+    p = m->attributes->pdata[i];
+    g_free(owl_pair_get_value(p));
+    g_slice_free(owl_pair, p);
   }
 
-  owl_list_free_simple(&(m->attributes));
+  g_ptr_array_free(m->attributes, true);
  
   owl_message_invalidate_format(m);
 }
+
+void owl_message_delete(owl_message *m)
+{
+  owl_message_cleanup(m);
+  g_slice_free(owl_message, m);
+}
