@@ -1,4 +1,4 @@
-/*  Copyright (c) 2006-2008 The BarnOwl Developers. All rights reserved.
+/*  Copyright (c) 2006-2011 The BarnOwl Developers. All rights reserved.
  *  Copyright (c) 2004 James Kretchmar. All rights reserved.
  *
  *  This program is free software. You can redistribute it and/or
@@ -6,23 +6,37 @@
  *  file included with the distribution for more information.
  */
 
-#ifndef INC_OWL_H
-#define INC_OWL_H
+#ifndef INC_BARNOWL_OWL_H
+#define INC_BARNOWL_OWL_H
+
+#include <config.h>
+
+#include "compat/compat.h"
+
+#include <stdbool.h>
 
 #ifndef OWL_PERL
-#include <curses.h>
+#define NCURSES_ENABLE_STDBOOL_H 1
+#include <ncursesw/curses.h>
+#include <ncursesw/panel.h>
 #endif
 #include <sys/param.h>
+#include <sys/types.h>
+#include <ctype.h>
+#include <errno.h>
 #include <EXTERN.h>
+#include <fcntl.h>
 #include <netdb.h>
 #include <regex.h>
 #include <time.h>
 #include <signal.h>
+#include <stdlib.h>
+#include <string.h>
 #include <termios.h>
-#include <libfaim/aim.h>
+#include <unistd.h>
+#include "libfaim/aim.h"
 #include <wchar.h>
-#include "config.h"
-#include "glib.h"
+#include <glib.h>
 #ifdef HAVE_LIBZEPHYR
 #include <zephyr/zephyr.h>
 #endif
@@ -32,41 +46,33 @@
 
 /* Perl and curses don't play nice. */
 #ifdef OWL_PERL
-typedef void WINDOW;
+typedef struct _owl_fake_WINDOW WINDOW;
+typedef struct _owl_fake_PANEL PANEL;
 /* logout is defined in FreeBSD. */
 #define logout logout_
-/* aim.h defines bool */
 #define HAS_BOOL
 #include <perl.h>
 #include "owl_perl.h"
 #undef logout
 #include "XSUB.h"
 #else
-typedef void SV;
+typedef struct _owl_fake_SV SV;
+typedef struct _owl_fake_AV AV;
+typedef struct _owl_fake_HV HV;
 #endif
 
-static const char owl_h_fileIdent[] = "$Id$";
+#include "window.h"
 
-#define BARNOWL_STRINGIFY(x) _STRINGIFY(x)
-#define _STRINGIFY(x) #x
-
-#ifndef OWL_SVN_REVNO
-#define OWL_SVN_REVNO ????
-#endif
-
-#ifndef OWL_VERSION_STRING
-#define OWL_VERSION_STRING "1.0.1"
-#endif
+extern const char *version;
 
 /* Feature that is being tested to redirect stderr through a pipe. 
  * There may still be some portability problems with this. */
 #define OWL_STDERR_REDIR 1
 
 #define OWL_DEBUG 0
-#define OWL_DEBUG_FILE "/var/tmp/owldebug"
+#define OWL_DEBUG_FILE "/var/tmp/barnowl-debug"
 
 #define OWL_CONFIG_DIR "/.owl"             /* this is relative to the user's home directory */
-#define OWL_STARTUP_FILE "/.owl/startup"   /* this is relative to the user's home directory */
 
 #define OWL_FMTEXT_ATTR_NONE      0
 #define OWL_FMTEXT_ATTR_BOLD      1
@@ -86,10 +92,6 @@ static const char owl_h_fileIdent[] = "$Id$";
 #define OWL_FMTEXT_UC_ALLCOLOR_MASK ( OWL_FMTEXT_UC_COLOR_MASK | OWL_FMTEXT_UC_DEFAULT_COLOR | 0x200)
 #define OWL_FMTEXT_UC_STARTBYTE_UTF8 '\xf4'
 
-#define OWL_FMTEXT_UTF8_ATTR_NONE "\xf4\x80\xa0\x80"
-#define OWL_FMTEXT_UTF8_FGDEFAULT "\xf4\x80\x94\x80"
-#define OWL_FMTEXT_UTF8_BGDEFAULT "\xf4\x80\x9C\x80"
-
 #define OWL_COLOR_BLACK     0
 #define OWL_COLOR_RED       1
 #define OWL_COLOR_GREEN     2
@@ -99,6 +101,9 @@ static const char owl_h_fileIdent[] = "$Id$";
 #define OWL_COLOR_CYAN      6
 #define OWL_COLOR_WHITE     7
 #define OWL_COLOR_DEFAULT   -1
+#define OWL_COLOR_INVALID   -2
+
+#define OWL_TAB_WIDTH 8
 
 #define OWL_EDITWIN_STYLE_MULTILINE 0
 #define OWL_EDITWIN_STYLE_ONELINE   1
@@ -114,10 +119,6 @@ static const char owl_h_fileIdent[] = "$Id$";
 #define OWL_MESSAGE_DIRECTION_IN    1
 #define OWL_MESSAGE_DIRECTION_OUT   2
 
-#define OWL_MUX_READ   1
-#define OWL_MUX_WRITE  2
-#define OWL_MUX_EXCEPT 4
-
 #define OWL_DIRECTION_NONE      0
 #define OWL_DIRECTION_DOWNWARDS 1
 #define OWL_DIRECTION_UPWARDS   2
@@ -158,106 +159,115 @@ static const char owl_h_fileIdent[] = "$Id$";
 #define OWL_CTX_EDITMULTI    0x2000
 #define OWL_CTX_EDITRESPONSE 0x4000
 
-#define OWL_USERCLUE_NONE       0
-#define OWL_USERCLUE_CLASSES    1
-#define OWL_USERCLUE_FOOBAR     2
-#define OWL_USERCLUE_BAZ        4
-
-#define OWL_WEBBROWSER_NONE     0
-#define OWL_WEBBROWSER_NETSCAPE 1
-#define OWL_WEBBROWSER_GALEON   2
-#define OWL_WEBBROWSER_OPERA    3
-
 #define OWL_VARIABLE_OTHER      0
 #define OWL_VARIABLE_INT        1
 #define OWL_VARIABLE_BOOL       2
 #define OWL_VARIABLE_STRING     3
 
+#define OWL_OUTPUT_RETURN       0
+#define OWL_OUTPUT_POPUP        1
+#define OWL_OUTPUT_ADMINMSG     2
+
 #define OWL_FILTER_MAX_DEPTH    300
 
 #define OWL_KEYMAP_MAXSTACK     20
 
+#define OWL_KEYBINDING_NOOP     0   /* dummy binding that does nothing */
 #define OWL_KEYBINDING_COMMAND  1   /* command string */
 #define OWL_KEYBINDING_FUNCTION 2   /* function taking no args */
 
 #define OWL_DEFAULT_ZAWAYMSG    "I'm sorry, but I am currently away from the terminal and am\nnot able to receive your message.\n"
 #define OWL_DEFAULT_AAWAYMSG    "I'm sorry, but I am currently away from the terminal and am\nnot able to receive your message.\n"
 
-#define OWL_INCLUDE_REG_TESTS   1  /* whether to build in regression tests */
-
 #define OWL_CMD_ALIAS_SUMMARY_PREFIX "command alias to: "
 
-#define OWL_WEBZEPHYR_PRINCIPAL "daemon.webzephyr"
+#define OWL_WEBZEPHYR_PRINCIPAL "daemon/webzephyr.mit.edu"
 #define OWL_WEBZEPHYR_CLASS     "webzephyr"
 #define OWL_WEBZEPHYR_OPCODE    "webzephyr"
 
-#define OWL_REGEX_QUOTECHARS    "+*.?[]^\\${}()"
+#define OWL_ZEPHYR_NOSTRIP_HOST         "host/"
+#define OWL_ZEPHYR_NOSTRIP_RCMD         "rcmd."
+#define OWL_ZEPHYR_NOSTRIP_DAEMON5      "daemon/"
+#define OWL_ZEPHYR_NOSTRIP_DAEMON4      "daemon."
+
+#define OWL_REGEX_QUOTECHARS    "!+*.?[]^\\${}()|"
 #define OWL_REGEX_QUOTEWITH     "\\"
 
 #if defined(HAVE_DES_STRING_TO_KEY) && defined(HAVE_DES_KEY_SCHED) && defined(HAVE_DES_ECB_ENCRYPT)
 #define OWL_ENABLE_ZCRYPT 1
 #endif
 
+/* Annotate functions in which the caller owns the return value and is
+ * responsible for ensuring it is freed. */
+#define CALLER_OWN G_GNUC_WARN_UNUSED_RESULT
+
 #define OWL_META(key) ((key)|010000)
 /* OWL_CTRL is definied in kepress.c */
 
-#define LINE 2048
+#ifdef HAVE_LIBZEPHYR
+/* libzephyr doesn't use const, so we appease the type system with this kludge.
+ * This just casts const char * to char * in a way that doesn't yield a warning
+ * from gcc -Wcast-qual. */
+static inline char *zstr(const char *str)
+{
+  union { char *rw; const char *ro; } u;
+  u.ro = str;
+  return u.rw;
+}
+#endif
+
+/* Convert char *const * into const char *const *.  This conversion is safe,
+ * and implicit in C++ (conv.qual 4) but for some reason not in C. */
+static inline const char *const *strs(char *const *pstr)
+{
+  return (const char *const *)pstr;
+}
 
 typedef struct _owl_variable {
   char *name;
   int   type;  /* OWL_VARIABLE_* */
-  void *pval_default;  /* for types other and string */
-  int   ival_default;  /* for types int and bool     */
-  char *validsettings;		/* documentation of valid settings */
+  char *default_str;            /* the default value as a string */
+  char *validsettings;	        /* documentation of valid settings */
   char *summary;		/* summary of usage */
   char *description;		/* detailed description */
-  void *val;                    /* current value */
-  int  (*validate_fn)(struct _owl_variable *v, void *newval);
+  bool takes_on_off;            /* allow passing on/off in argument-less set/unset */
+  GClosure *set_fromstring_fn;
+                                /* sets the variable to a value
+                                 * of the appropriate type.
+                                 * unless documented, this
+                                 * should make a copy.
+                                 * returns 0 on success. */
+  GClosure *get_tostring_fn;
+                                /* converts val to a string;
+                                 * caller must free the result */
+
+  /* These are only valid for OWL_VARIABLE_{INT,BOOL,STRING} */
+  GValue val;                   /* current value, if default get_fn/set_fn */
+
+  GCallback get_fn;
+                                /* returns a reference to the current value.
+                                 * WARNING:  this approach is hard to make
+                                 * thread-safe... */
+  GCallback validate_fn;
                                 /* returns 1 if newval is valid */
-  int  (*set_fn)(struct _owl_variable *v, void *newval); 
+  GCallback set_fn;
                                 /* sets the variable to a value
-				 * of the appropriate type.
-				 * unless documented, this 
-				 * should make a copy. 
-				 * returns 0 on success. */
-  int  (*set_fromstring_fn)(struct _owl_variable *v, char *newval);
-                                /* sets the variable to a value
-				 * of the appropriate type.
-				 * unless documented, this 
-				 * should make a copy. 
-				 * returns 0 on success. */
-  void *(*get_fn)(struct _owl_variable *v);
-				/* returns a reference to the current value.
-				 * WARNING:  this approach is hard to make
-				 * thread-safe... */
-  int  (*get_tostring_fn)(struct _owl_variable *v, 
-			  char *buf, int bufsize, void *val); 
-                                /* converts val to a string 
-				 * and puts into buf */
-  void  (*free_fn)(struct _owl_variable *v);
-				/* frees val as needed */
+                                 * of the appropriate type.
+                                 * unless documented, this
+                                 * should make a copy.
+                                 * returns 0 on success. */
 } owl_variable;
 
+
 typedef struct _owl_input {
   int ch;
   gunichar uch;
 } owl_input;
 
 typedef struct _owl_fmtext {
-  int textlen;
-  int bufflen;
-  char *textbuff;
-  char default_attrs;
-  short default_fgcolor;
-  short default_bgcolor;
+  GString *buff;
 } owl_fmtext;
 
-typedef struct _owl_list {
-  int size;
-  int avail;
-  void **list;
-} owl_list;
-
 typedef struct _owl_dict_el {
   char *k;			/* key   */
   void *v;			/* value */
@@ -274,6 +284,11 @@ typedef owl_dict owl_cmddict;	/* dict of commands */
 typedef struct _owl_context {
   int   mode;
   void *data;		/* determined by mode */
+  char *keymap;
+  owl_window *cursor;
+  void (*deactivate_cb)(struct _owl_context*);
+  void (*delete_cb)(struct _owl_context*);
+  void *cbdata;
 } owl_context;
 
 typedef struct _owl_cmd {	/* command */
@@ -293,14 +308,14 @@ typedef struct _owl_cmd {	/* command */
   char *cmd_aliased_to;		/* what this command is aliased to... */
   
   /* These don't take any context */
-  char *(*cmd_args_fn)(int argc, char **argv, char *buff);  
+  CALLER_OWN char *(*cmd_args_fn)(int argc, const char *const *argv, const char *buff);
 				/* takes argv and the full command as buff.
 				 * caller must free return value if !NULL */
   void (*cmd_v_fn)(void);	/* takes no args */
   void (*cmd_i_fn)(int i);	/* takes an int as an arg */
 
   /* The following also take the active context if it's valid */
-  char *(*cmd_ctxargs_fn)(void *ctx, int argc, char **argv, char *buff);  
+  CALLER_OWN char *(*cmd_ctxargs_fn)(void *ctx, int argc, const char *const *argv, const char *buff);
 				/* takes argv and the full command as buff.
 				 * caller must free return value if !NULL */
   void (*cmd_ctxv_fn)(void *ctx);	        /* takes no args */
@@ -310,20 +325,22 @@ typedef struct _owl_cmd {	/* command */
 
 
 typedef struct _owl_zwrite {
+  char *cmd;
+  char *zwriteline;
   char *class;
   char *inst;
   char *realm;
   char *opcode;
   char *zsig;
   char *message;
-  owl_list recips;
+  GPtrArray *recips;
   int cc;
   int noping;
 } owl_zwrite;
 
 typedef struct _owl_pair {
-  void *key;
-  void *value;
+  const char *key;
+  char *value;
 } owl_pair;
 
 struct _owl_fmtext_cache;
@@ -332,15 +349,15 @@ typedef struct _owl_message {
   int id;
   int direction;
 #ifdef HAVE_LIBZEPHYR
+  bool has_notice;
   ZNotice_t notice;
 #endif
   struct _owl_fmtext_cache * fmtext;
   int delete;
-  char *hostname;
-  owl_list attributes;            /* this is a list of pairs */
+  const char *hostname;
+  GPtrArray *attributes;          /* this is a list of pairs */
   char *timestr;
   time_t time;
-  char *zwriteline;
 } owl_message;
 
 #define OWL_FMTEXT_CACHE_SIZE 1000
@@ -360,39 +377,44 @@ typedef struct _owl_mainwin {
   int curtruncated;
   int lasttruncated;
   int lastdisplayed;
+  owl_window *window;
 } owl_mainwin;
 
+typedef struct _owl_editwin owl_editwin;
+typedef struct _owl_editwin_excursion owl_editwin_excursion;
+
 typedef struct _owl_viewwin {
   owl_fmtext fmtext;
   int textlines;
   int topline;
   int rightshift;
-  int winlines, wincols;
-  WINDOW *curswin;
+  owl_window *window;
   void (*onclose_hook) (struct _owl_viewwin *vwin, void *data);
   void *onclose_hook_data;
+
+  gulong sig_resize_id;
+  owl_window *content;
+  gulong sig_content_redraw_id;
+  owl_window *status;
+  gulong sig_status_redraw_id;
+  owl_window *cmdwin;
 } owl_viewwin;
   
 typedef struct _owl_popwin {
-  WINDOW *borderwin;
-  WINDOW *popwin;
-  int lines;
-  int cols;
-  int active;
-  int needsfirstrefresh;
-  void (*handler) (int ch);
+  owl_window *border;
+  owl_window *content;
+  gulong sig_redraw_id;
+  gulong sig_resize_id;
 } owl_popwin;
-
-typedef struct _owl_popexec {
-  int refcount;
-  owl_viewwin *vwin;
-  int winactive;
-  int pid;			/* or 0 if it has terminated */
-  int rfd;  
-} owl_popexec;
+  
+typedef struct _owl_msgwin {
+  char *msg;
+  owl_window *window;
+  gulong redraw_id;
+} owl_msgwin;
 
 typedef struct _owl_messagelist {
-  owl_list list;
+  GPtrArray *list;
 } owl_messagelist;
 
 typedef struct _owl_regex {
@@ -402,9 +424,9 @@ typedef struct _owl_regex {
 } owl_regex;
 
 typedef struct _owl_filterelement {
-  int (*match_message)(struct _owl_filterelement *fe, owl_message *m);
+  int (*match_message)(const struct _owl_filterelement *fe, const owl_message *m);
   /* Append a string representation of the filterelement onto buf*/
-  void (*print_elt)(struct _owl_filterelement *fe, char * buf);
+  void (*print_elt)(const struct _owl_filterelement *fe, GString *buf);
   /* Operands for and,or,not*/
   struct _owl_filterelement *left, *right;
   /* For regex filters*/
@@ -415,59 +437,37 @@ typedef struct _owl_filterelement {
 
 typedef struct _owl_filter {
   char *name;
-  int polarity;
   owl_filterelement * root;
   int fgcolor;
   int bgcolor;
-  int cachedmsgid;  /* cached msgid: should move into view eventually */
 } owl_filter;
 
 typedef struct _owl_view {
   char *name;
   owl_filter *filter;
-  owl_messagelist ml;
-  owl_style *style;
+  owl_messagelist *ml;
+  const owl_style *style;
+  int cachedmsgid;
 } owl_view;
 
 typedef struct _owl_history {
-  owl_list hist;
-  int cur;
-  int touched;
-  int partial;
-  int repeats;
+  GQueue hist;
+  GList *cur;
+  bool partial;
 } owl_history;
 
-typedef struct _owl_editwin {
-  char *buff;
-  owl_history *hist;
-  int bufflen;
-  int allocated;
-  int buffx, buffy;
-  int topline;
-  int winlines, wincols, fillcol, wrapcol;
-  WINDOW *curswin;
-  int style;
-  int lock;
-  int dotsend;
-  int echochar;
-
-  char *command;
-  void (*callback)(struct _owl_editwin*);
-  void *cbdata;
-} owl_editwin;
-
-typedef struct _owl_mux {
-  int handle;			/* for referencing this */
-  int active;			/* has this been deleted? */
-  int fd;		       
-  int eventmask;		/* bitmask of OWL_MUX_* */
-  void (*handler_fn)(int handle, int fd, int eventmask, void *data);
-  void *data;			/* data reference to pass to callback */
-} owl_mux;
-typedef owl_list owl_muxevents;
+typedef struct _owl_mainpanel {
+  owl_window *panel;
+  owl_window *typwin;
+  owl_window *sepwin;
+  owl_window *msgwin;
+  owl_window *recwin;
+  int recwinlines;
+} owl_mainpanel;
 
 typedef struct _owl_keybinding {
-  int  *j;			/* keypress stack (0-terminated) */  
+  int  *keys;			/* keypress stack */
+  int   len;                    /* length of stack */
   int   type;			/* command or function? */
   char *desc;			/* description (or "*user*") */
   char *command;		/* command, if of type command */
@@ -477,8 +477,8 @@ typedef struct _owl_keybinding {
 typedef struct _owl_keymap {
   char     *name;		/* name of keymap */
   char     *desc;		/* description */
-  owl_list  bindings;		/* key bindings */
-  struct _owl_keymap *submap;	/* submap */
+  GPtrArray *bindings;		/* key bindings */
+  const struct _owl_keymap *parent;	/* parent */
   void (*default_fn)(owl_input j);	/* default action (takes a keypress) */
   void (*prealways_fn)(owl_input  j);	/* always called before a keypress is received */
   void (*postalways_fn)(owl_input  j);	/* always called after keypress is processed */
@@ -486,7 +486,7 @@ typedef struct _owl_keymap {
 
 typedef struct _owl_keyhandler {
   owl_dict  keymaps;		/* dictionary of keymaps */
-  owl_keymap *active;		/* currently active keymap */
+  const owl_keymap *active;		/* currently active keymap */
   int	    in_esc;		/* escape pressed? */
   int       kpstack[OWL_KEYMAP_MAXSTACK+1]; /* current stack of keypresses */
   int       kpstackpos;		/* location in stack (-1 = none) */
@@ -500,121 +500,118 @@ typedef struct _owl_buddy {
 } owl_buddy;
 
 typedef struct _owl_buddylist {
-  owl_list buddies;
+  GPtrArray *buddies;
 } owl_buddylist;
 
 typedef struct _owl_zbuddylist {
-  owl_list zusers;
+  GPtrArray *zusers;
 } owl_zbuddylist;
 
-typedef struct _owl_timer {
-  int direction;
-  time_t starttime;
-  int start;
-} owl_timer;
-
 typedef struct _owl_errqueue {
-  owl_list errlist;
+  GPtrArray *errlist;
 } owl_errqueue;
 
 typedef struct _owl_colorpair_mgr {
   int next;
   short **pairs;
+  bool overflow;
 } owl_colorpair_mgr;
 
-typedef struct _owl_obarray {
-  owl_list strings;
-} owl_obarray;
-
-typedef struct _owl_dispatch {
-  int fd;           /* FD to watch for dispatch. */
-  void (*cfunc)();  /* C function to dispatch to. */
-  SV *pfunc;        /* Perl function to dispatch to. */
-} owl_dispatch;
+typedef struct _owl_popexec {
+  int refcount;
+  owl_viewwin *vwin;
+  int winactive;
+  pid_t pid;			/* or 0 if it has terminated */
+  guint io_watch;
+} owl_popexec;
 
 typedef struct _owl_global {
-  owl_mainwin mw;
-  owl_popwin pw;
+  owl_mainwin *mw;
+  owl_popwin *pw;
+  owl_msgwin msgwin;
   owl_history cmdhist;		/* command history */
   owl_history msghist;		/* outgoing message history */
   owl_keyhandler kh;
-  owl_list filterlist;
-  owl_list puntlist;
-  owl_muxevents muxevents;	/* fds to dispatch on */
+  owl_dict filters;
+  GList *filterlist;
+  GPtrArray *puntlist;
   owl_vardict vars;
   owl_cmddict cmds;
-  owl_context ctx;
+  GList *context_stack;
   owl_errqueue errqueue;
   int lines, cols;
   int curmsg, topmsg;
+  int markedmsgid;              /* for finding the marked message when it has moved. */
   int curmsg_vert_offset;
   owl_view current_view;
-  owl_messagelist msglist;
-  WINDOW *recwin, *sepwin, *msgwin, *typwin;
-  int needrefresh;
+  owl_messagelist *msglist;
+  WINDOW *input_pad;
+  owl_mainpanel mainpanel;
+  gulong typwin_erase_id;
   int rightshift;
-  int resizepending;
-  int recwinlines;
-  int typwinactive;
-  char *thishost;
+  bool resizepending;
   char *homedir;
   char *confdir;
   char *startupfile;
   int direction;
-  int zaway;
-  char *cur_zaway_msg;
   int haveconfig;
   int config_format;
-  void *buffercbdata;
-  owl_editwin tw;
-  owl_viewwin vw;
+  owl_editwin *tw;
+  owl_viewwin *vw;
   void *perl;
   int debug;
   time_t starttime;
   time_t lastinputtime;
+  time_t last_wakeup_time;
   char *startupargs;
-  int userclue;
   int nextmsgid;
-  int hascolors;
-  int colorpairs;
   owl_colorpair_mgr cpmgr;
-  int searchactive;
-  int newmsgproc_pid;
-  int malloced, freed;
-  char *searchstring;
+  pid_t newmsgproc_pid;
+  owl_regex search_re;
   aim_session_t aimsess;
   aim_conn_t bosconn;
-  owl_timer aim_noop_timer;
-  owl_timer aim_ignorelogin_timer;
-  owl_timer aim_buddyinfo_timer;
   int aim_loggedin;         /* true if currently logged into AIM */
-  int aim_doprocessing;     /* true if we should process AIM events (like pending login) */
+  GSource *aim_event_source; /* where we get our AIM events from */
   char *aim_screenname;     /* currently logged in AIM screen name */
   char *aim_screenname_for_filters;     /* currently logged in AIM screen name */
   owl_buddylist buddylist;  /* list of logged in AIM buddies */
-  owl_list messagequeue;    /* for queueing up aim and other messages */
+  GQueue *messagequeue;     /* for queueing up aim and other messages */
   owl_dict styledict;       /* global dictionary of available styles */
   char *response;           /* response to the last question asked */
   int havezephyr;
   int haveaim;
-  int got_err_signal;	    /* 1 if we got an unexpected signal */
-  siginfo_t err_signal_info;
+  int ignoreaimlogin;
   owl_zbuddylist zbuddies;
-  owl_timer zephyr_buddycheck_timer;
+  GList *zaldlist;
+  int pseudologin_notify;
   struct termios startup_tio;
-  owl_obarray obarray;
-  owl_list dispatchlist;
+  guint aim_nop_timer;
+  int load_initial_subs;
+  FILE *debug_file;
+  char *kill_buffer;
+  int interrupt_count;
+#if GLIB_CHECK_VERSION(2, 31, 0)
+  GMutex interrupt_lock;
+#else
+  GMutex *interrupt_lock;
+#endif
 } owl_global;
 
 /* globals */
 extern owl_global g;
 
-#include "owl_prototypes.h"
+#include <owl_prototypes.h>
 
-/* these are missing from the zephyr includes for some reason */
-#ifdef HAVE_LIBZEPHYR
+/* These were missing from the Zephyr includes before Zephyr 3. */
+#if defined HAVE_LIBZEPHYR && defined ZCONST
 int ZGetSubscriptions(ZSubscription_t *, int *);
 int ZGetLocations(ZLocations_t *,int *);
 #endif
 
-#endif /* INC_OWL_H */
+/* We have to dynamically bind these ourselves */
+extern gboolean (*gvalue_from_sv) (GValue * value, SV * sv);
+extern SV * (*sv_from_gvalue) (const GValue * value);
+extern GClosure * (*perl_closure_new) (SV * callback, SV * data, gboolean swap);
+
+
+#endif /* INC_BARNOWL_OWL_H */
