@@ -7,7 +7,7 @@ Module::Install::BarnOwl
 
 =head1 DESCRIPTION
 
-Module::Install::BarnOwl is a M::I module to help building barnowl
+Module::Install::BarnOwl is a M::I module to help building BarnOwl
 modules,
 
 =head1 SYNOPSIS
@@ -27,7 +27,7 @@ This is roughly equivalent to:
     WriteAll;
 
 As well as make rules to generate Jabber.par, and to put some
-additional barnowl-specific information into META.yml
+additional BarnOwl-specific information into META.yml
 
 =cut
 
@@ -40,15 +40,22 @@ sub barnowl_module {
     my $name = ucfirst shift;
     my $class = ref $self;
 
+    my $srcdir = $ENV{BARNOWL_SRCDIR} || '.';
+
     $self->name("BarnOwl-Module-$name");
-    $self->all_from("lib/BarnOwl/Module/$name.pm");
+    $self->all_from("$srcdir/lib/BarnOwl/Module/$name.pm");
+    $self->makemaker_args(PMLIBDIRS => ["$srcdir/lib"],
+                          PMLIBPARENTDIRS => ["$srcdir/lib"]);
 
     $self->postamble(<<"END_MAKEFILE");
 
 # --- $class section:
 
+BARNOWL_SRCDIR = $srcdir
+export BARNOWL_SRCDIR
+
 $name.par: pm_to_blib
-\tcd blib; zip ../$name.par -r arch lib
+\tcd blib; zip -q ../$name.par -r arch lib
 
 END_MAKEFILE
 }
