@@ -9,7 +9,7 @@ BarnOwl::Module::Jabber
 
 =head1 DESCRIPTION
 
-This module implements Jabber support for barnowl.
+This module implements Jabber support for BarnOwl.
 
 =cut
 
@@ -18,6 +18,7 @@ use BarnOwl::Hooks;
 use BarnOwl::Message::Jabber;
 use BarnOwl::Module::Jabber::Connection;
 use BarnOwl::Module::Jabber::ConnectionManager;
+use BarnOwl::Completion::Util qw(complete_flags);
 
 use Authen::SASL qw(Perl);
 use Net::Jabber;
@@ -34,8 +35,8 @@ BEGIN {
     if(eval {require IO::Socket::SSL;}) {
         if($IO::Socket::SSL::VERSION eq "0.97") {
             BarnOwl::error("You are using IO::Socket:SSL 0.97, which \n" .
-                           "contains bugs causing it not to work with barnowl's\n" .
-                           "jabber support. We recommend updating to the latest\n" .
+                           "contains bugs causing it not to work with BarnOwl's\n" .
+                           "Jabber support. We recommend updating to the latest\n" .
                            "IO::Socket::SSL from CPAN. \n");
             die("Not loading Jabber.par\n");
         }
@@ -61,18 +62,22 @@ no warnings 'redefine';
 our $conn;
 $conn ||= BarnOwl::Module::Jabber::ConnectionManager->new;
 our %vars;
+our %completion_jids;
 
 sub onStart {
     if ( *BarnOwl::queue_message{CODE} ) {
         register_owl_commands();
         register_keybindings();
         register_filters();
-        $BarnOwl::Hooks::mainLoop->add(\&onMainLoop);
-        $BarnOwl::Hooks::getBuddyList->add(\&onGetBuddyList);
+        $BarnOwl::Hooks::getBuddyList->add("BarnOwl::Module::Jabber::onGetBuddyList");
+        $BarnOwl::Hooks::getQuickstart->add("BarnOwl::Module::Jabber::onGetQuickstart");
         $vars{show} = '';
 	BarnOwl::new_variable_bool("jabber:show_offline_buddies",
 				   { default => 1,
 				     summary => 'Show offline or pending buddies.'});
+	BarnOwl::new_variable_bool("jabber:show_logins",
+				   { default => 0,
+				     summary => 'Show login/logout messages.'});
 	BarnOwl::new_variable_bool("jabber:spew",
 				   { default => 0,
 				     summary => 'Display unrecognized Jabber messages.'});
@@ -84,6 +89,16 @@ sub onStart {
 				  { default => 15,
 				    summary => 'After minutes idle, auto extended away.'
 				});
+	BarnOwl::new_variable_bool("jabber:reconnect",
+				  { default => 1,
+				    summary => 'Auto-reconnect when disconnected from servers.'
+				});
+        # Force these. Reload can screw them up.
+        # Taken from Net::Jabber::Protocol.
+        $Net::XMPP::Protocol::NEWOBJECT{'iq'}       = "Net::Jabber::IQ";
+        $Net::XMPP::Protocol::NEWOBJECT{'message'}  = "Net::Jabber::Message";
+        $Net::XMPP::Protocol::NEWOBJECT{'presence'} = "Net::Jabber::Presence";
+        $Net::XMPP::Protocol::NEWOBJECT{'jid'}      = "Net::Jabber::JID";
     } else {
         # Our owl doesn't support queue_message. Unfortunately, this
         # means it probably *also* doesn't support BarnOwl::error. So just
@@ -91,10 +106,17 @@ sub onStart {
     }
 }
 
-$BarnOwl::Hooks::startup->add(\&onStart);
+$BarnOwl::Hooks::startup->add("BarnOwl::Module::Jabber::onStart");
 
-sub onMainLoop {
-    return if ( !$conn->connected() );
+sub do_keep_alive_and_auto_away {
+    if ( !$conn->connected() ) {
+        # We don't need this timer any more.
+        if (defined $vars{keepAliveTimer}) {
+            $vars{keepAliveTimer}->stop;
+        }
+        delete $vars{keepAliveTimer};
+        return;
+    }
 
     $vars{status_changed} = 0;
     my $auto_away = BarnOwl::getvar('jabber:auto_away_timeout');
@@ -108,24 +130,29 @@ sub onMainLoop {
         $vars{show} = 'away';
         $vars{status} = 'Auto away after '.$auto_away.' minute'.($auto_away == 1 ? '' : 's').' idle.';
         $vars{status_changed} = 1;
-    } elsif ($idletime == 0 && $vars{show} ne '') {
+    } elsif ($idletime <= $vars{idletime} && $vars{show} ne '') {
         $vars{show} = '';
         $vars{status} = '';
         $vars{status_changed} = 1;
     }
+    $vars{idletime} = $idletime;
 
     foreach my $jid ( $conn->getJIDs() ) {
-        my $client = $conn->getConnectionFromJID($jid);
+        next if $conn->jidActive($jid);
+        $conn->tryReconnect($jid);
+    }
 
+    foreach my $jid ( $conn->getJIDs() ) {
+        next unless $conn->jidActive($jid);
+
+        my $client = $conn->getConnectionFromJID($jid);
         unless($client) {
             $conn->removeConnection($jid);
             BarnOwl::error("Connection for $jid undefined -- error in reload?");
         }
-        # We keep this in the mainloop hook for keep-alives
-        my $status = $client->Process(0);
+        my $status = $client->Process(0); # keep-alive
         if ( !defined($status) ) {
-            BarnOwl::error("Jabber account $jid disconnected!");
-            do_logout($jid);
+            $conn->scheduleReconnect($jid);
         }
         if ($::shutdown) {
             do_logout($jid);
@@ -153,6 +180,8 @@ sub blist_listBuddy {
     my $name = $jq{name} || $buddy->GetUserID();
 
     $blistStr .= sprintf '%-15s %s', $name, $buddy->GetJID();
+    $completion_jids{$name} = 1;
+    $completion_jids{$buddy->GetJID()} = 1;
 
     if ($res) {
         my %rq = $roster->resourceQuery( $buddy, $res );
@@ -175,6 +204,11 @@ sub blist_listBuddy {
     return $blistStr . "\n";
 }
 
+# Sort, ignoring markup.
+sub blistSort {
+    return uc(BarnOwl::ztext_stylestrip($a)) cmp uc(BarnOwl::ztext_stylestrip($b));
+}
+
 sub getSingleBuddyList {
     my $jid = shift;
     $jid = resolveConnectedJID($jid);
@@ -182,23 +216,29 @@ sub getSingleBuddyList {
     my $blist = "";
     my $roster = $conn->getRosterFromJID($jid);
     if ($roster) {
-        $blist .= "\n" . BarnOwl::Style::boldify("Jabber Roster for $jid\n");
+        $blist .= BarnOwl::Style::boldify("Jabber roster for $jid\n");
 
+        my @gTexts = ();
         foreach my $group ( $roster->groups() ) {
-            $blist .= "  Group: $group\n";
             my @buddies = $roster->jids( 'group', $group );
+            my @bTexts = ();
             foreach my $buddy ( @buddies ) {
-                $blist .= blist_listBuddy( $roster, $buddy );
+                push(@bTexts, blist_listBuddy( $roster, $buddy ));
             }
+            push(@gTexts, "  Group: $group\n".join('',sort blistSort @bTexts));
         }
+        # Sort groups before adding ungrouped entries.
+        @gTexts = sort blistSort @gTexts;
 
         my @unsorted = $roster->jids('nogroup');
         if (@unsorted) {
-            $blist .= "  [unsorted]\n";
+            my @bTexts = ();
             foreach my $buddy (@unsorted) {
-                $blist .= blist_listBuddy( $roster, $buddy );
+                push(@bTexts, blist_listBuddy( $roster, $buddy ));
             }
+            push(@gTexts, "  [unsorted]\n".join('',sort blistSort @bTexts));
         }
+        $blist .= join('', @gTexts);
     }
     return $blist;
 }
@@ -212,25 +252,49 @@ sub onGetBuddyList {
     return $blist;
 }
 
+sub onGetQuickstart {
+    return <<'EOF'
+@b(Jabber:)
+Type ':jabberlogin @b(username@mit.edu)' to log in to Jabber. The command
+':jroster sub @b(somebody@gmail.com)' will request that they let you message
+them. Once you get a message saying you are subscribed, you can message
+them by typing ':jwrite @b(somebody@gmail.com)' or just 'j @b(somebody)'.
+EOF
+}
+
 ################################################################################
 ### Owl Commands
 sub register_owl_commands() {
     BarnOwl::new_command(
         jabberlogin => \&cmd_login,
         {
-            summary => "Log into jabber",
-            usage   => "jabberlogin JID [PASSWORD]"
+            summary => "Log in to Jabber",
+            usage   => "jabberlogin <jid> [<password>]"
         }
     );
     BarnOwl::new_command(
         jabberlogout => \&cmd_logout,
-        { summary => "Log out of jabber" }
+        {
+            summary => "Log out of Jabber",
+            usage   => "jabberlogout [-A|<jid>]",
+            description => "jabberlogout logs you out of Jabber.\n\n"
+              . "If you are connected to one account, no further arguments are necessary.\n\n"
+              . "-A            Log out of all accounts.\n"
+              . "<jid>         Which account to log out of.\n"
+        }
     );
     BarnOwl::new_command(
         jwrite => \&cmd_jwrite,
         {
             summary => "Send a Jabber Message",
-            usage   => "jwrite JID [-t thread] [-s subject]"
+            usage   => "jwrite <jid> [-t <thread>] [-s <subject>] [-a <account>] [-m <message>]"
+        }
+    );
+    BarnOwl::new_command(
+        jaway => \&cmd_jaway,
+        {
+            summary => "Set Jabber away / presence information",
+            usage   => "jaway [-s online|dnd|...] [<message>]"
         }
     );
     BarnOwl::new_command(
@@ -244,37 +308,38 @@ sub register_owl_commands() {
         jmuc => \&cmd_jmuc,
         {
             summary     => "Jabber MUC related commands.",
-            description => "jmuc sends jabber commands related to muc.\n\n"
+            description => "jmuc sends Jabber commands related to MUC.\n\n"
               . "The following commands are available\n\n"
-              . "join <muc>  Join a muc.\n\n"
-              . "part <muc>  Part a muc.\n"
-              . "            The muc is taken from the current message if not supplied.\n\n"
-              . "invite <jid> <muc>\n"
+              . "join <muc>[/<nick>]\n"
+              . "            Join a MUC (with a given nickname, or otherwise your JID).\n\n"
+              . "part <muc>  Part a MUC.\n"
+              . "            The MUC is taken from the current message if not supplied.\n\n"
+              . "invite <jid> [<muc>]\n"
               . "            Invite <jid> to <muc>.\n"
-              . "            The muc is taken from the current message if not supplied.\n\n"
-              . "configure <muc>\n"
+              . "            The MUC is taken from the current message if not supplied.\n\n"
+              . "configure [<muc>]\n"
               . "            Configures a MUC.\n"
-              . "            Necessary to initalize a new MUC.\n"
+              . "            Necessary to initialize a new MUC.\n"
               . "            At present, only the default configuration is supported.\n"
-              . "            The muc is taken from the current message if not supplied.\n\n"
-              . "presence <muc>\n"
+              . "            The MUC is taken from the current message if not supplied.\n\n"
+              . "presence [<muc>]\n"
               . "            Shows the roster for <muc>.\n"
-              . "            The muc is taken from the current message if not supplied.\n\n"
+              . "            The MUC is taken from the current message if not supplied.\n\n"
               . "presence -a\n"
               . "            Shows rosters for all MUCs you're participating in.\n\n",
-            usage => "jmuc COMMAND ARGS"
+            usage => "jmuc <command> [<args>]"
         }
     );
     BarnOwl::new_command(
         jroster => \&cmd_jroster,
         {
-            summary     => "Jabber Roster related commands.",
-            description => "jroster sends jabber commands related to rosters.\n\n"
+            summary     => "Jabber roster related commands.",
+            description => "jroster sends Jabber commands related to rosters.\n\n"
               . "The following commands are available\n\n"
               . "sub <jid>     Subscribe to <jid>'s presence. (implicit add)\n\n"
               . "add <jid>     Adds <jid> to your roster.\n\n"
               . "unsub <jid>   Unsubscribe from <jid>'s presence.\n\n"
-              . "remove <jid>  Removes <jid> to your roster. (implicit unsub)\n\n"
+              . "remove <jid>  Removes <jid> from your roster. (implicit unsub)\n\n"
               . "auth <jid>    Authorizes <jid> to subscribe to your presence.\n\n"
               . "deauth <jid>  De-authorizes <jid>'s subscription to your presence.\n\n"
               . "The following arguments are supported for all commands\n\n"
@@ -284,43 +349,53 @@ sub register_owl_commands() {
               . "-g <group>    Add <jid> to group <group>.\n"
               . "              May be specified more than once, will not remove <jid> from any groups.\n\n"
               . "-p            Purge. Removes <jid> from all groups.\n"
-              . "              May be combined with -g completely alter <jid>'s groups.\n\n"
+              . "              May be combined with -g.\n\n"
               . "-n <name>     Sets <name> as <jid>'s short name.\n\n"
               . "Note: Unless -n is used, you can specify multiple <jid> arguments.\n",
-            usage       => "jroster COMMAND ARGS"
+            usage       => "jroster <command> <args>"
         }
     );
 }
 
 sub register_keybindings {
-    BarnOwl::bindkey("recv j command start-command jwrite ");
+    BarnOwl::bindkey(qw(recv j command start-command), 'jwrite ');
 }
 
 sub register_filters {
-    BarnOwl::filter('jabber type ^jabber$');
+    BarnOwl::filter(qw(jabber type ^jabber$));
 }
 
 sub cmd_login {
     my $cmd = shift;
+    my $jidStr = shift;
     my $jid = new Net::Jabber::JID;
-    $jid->SetJID(shift);
+    $jid->SetJID($jidStr);
     my $password = '';
     $password = shift if @_;
 
     my $uid           = $jid->GetUserID();
     my $componentname = $jid->GetServer();
-    my $resource      = $jid->GetResource() || 'owl';
-    $jid->SetResource($resource);
-    my $jidStr = $jid->GetJID('full');
+    my $resource      = $jid->GetResource();
 
-    if ( !$uid || !$componentname ) {
-        BarnOwl::error("usage: $cmd JID");
-        return;
+    if ($resource eq '') {
+        my $cjidStr = $conn->baseJIDExists($jidStr);
+        if ($cjidStr) {
+            die("Already logged in as $cjidStr.\n");
+        }
     }
 
-    if ( $conn->jidExists($jidStr) ) {
-        BarnOwl::error("Already logged in as $jidStr.");
-        return;
+    $resource ||= 'barnowl';
+    $jid->SetResource($resource);
+    $jidStr = $jid->GetJID('full');
+
+    if ( !$uid || !$componentname ) {
+        die("usage: $cmd JID\n");
+    }
+
+    if ( $conn->jidActive($jidStr) ) {
+        die("Already logged in as $jidStr.\n");
+    } elsif ($conn->jidExists($jidStr)) {
+        return $conn->tryReconnect($jidStr, 1);
     }
 
     my ( $server, $port ) = getServerFromJID($jid);
@@ -345,7 +420,7 @@ sub do_login {
     $vars{jlogin_authhash}->{password} = sub { return $vars{jlogin_password} || '' };
     my $jidStr = $vars{jlogin_jid};
     if ( !$jidStr && $vars{jlogin_havepass}) {
-        BarnOwl::error("Got password but have no jid!");
+        BarnOwl::error("Got password but have no JID!");
     }
     else
     {
@@ -374,7 +449,7 @@ sub do_login {
         my $status = $client->Connect( %{ $vars{jlogin_connhash} } );
         if ( !$status ) {
             $conn->removeConnection($jidStr);
-            BarnOwl::error("We failed to connect");
+            BarnOwl::error("We failed to connect.");
         } else {
             my @result = $client->AuthSend( %{ $vars{jlogin_authhash} } );
 
@@ -388,16 +463,13 @@ sub do_login {
                 $conn->removeConnection($jidStr);
                 BarnOwl::error( "Error in connect: " . join( " ", @result ) );
             } else {
-                $conn->getRosterFromJID($jidStr)->fetch();
-                $client->PresenceSend( priority => 1 );
-		my $fullJid = $client->{SESSION}->{FULLJID} || $jidStr;
-		$conn->renameConnection($jidStr, $fullJid);
-                queue_admin_msg("Connected to jabber as $fullJid");
-                # The remove_dispatch() method is called from the
-                # ConnectionManager's removeConnection() method.
-                $client->{fileno} = $client->getSocket()->fileno();
-                #queue_admin_msg("Connected to jabber as $fullJid ($client->{fileno})");
-                BarnOwl::add_dispatch($client->{fileno}, sub { $client->OwlProcess() });
+                $conn->setAuth(
+                    $jidStr,
+                    {   %{ $vars{jlogin_authhash} },
+                        password => $vars{jlogin_password}
+                    }
+                );
+                $client->onConnect($conn, $jidStr);
             }
         }
     }
@@ -418,6 +490,7 @@ sub do_logout {
 }
 
 sub cmd_logout {
+    return "You are not logged into Jabber." unless ($conn->connected() > 0);
     # Logged into multiple accounts
     if ( $conn->connected() > 1 ) {
         # Logged into multiple accounts, no accout specified.
@@ -431,7 +504,7 @@ sub cmd_logout {
         }
         # Logged into multiple accounts, account specified.
         else {
-            if ( $_[1] eq '-a' )    #All accounts.
+            if ( $_[1] eq '-A' )    #All accounts.
             {
                 foreach my $jid ( $conn->getJIDs() ) {
                     do_logout($jid);
@@ -453,16 +526,14 @@ sub cmd_logout {
 
 sub cmd_jlist {
     if ( !( scalar $conn->getJIDs() ) ) {
-        BarnOwl::error("You are not logged in to Jabber.");
-        return;
+        die("You are not logged in to Jabber.\n");
     }
     BarnOwl::popless_ztext( onGetBuddyList() );
 }
 
 sub cmd_jwrite {
     if ( !$conn->connected() ) {
-        BarnOwl::error("You are not logged in to Jabber.");
-        return;
+        die("You are not logged in to Jabber.\n");
     }
 
     my $jwrite_to      = "";
@@ -470,6 +541,7 @@ sub cmd_jwrite {
     my $jwrite_sid     = "";
     my $jwrite_thread  = "";
     my $jwrite_subject = "";
+    my $jwrite_body;
     my ($to, $from);
     my $jwrite_type    = "chat";
 
@@ -481,14 +553,13 @@ sub cmd_jwrite {
         'thread=s'  => \$jwrite_thread,
         'subject=s' => \$jwrite_subject,
         'account=s' => \$from,
-        'id=s'     =>  \$jwrite_sid,
-    ) or die("Usage: jwrite JID [-t thread] [-s 'subject'] [-a account]\n");
+        'id=s'      => \$jwrite_sid,
+        'message=s' => \$jwrite_body,
+    ) or die("Usage: jwrite <jid> [-t <thread>] [-s <subject>] [-a <account>]\n");
     $jwrite_type = 'groupchat' if $gc;
 
     if ( scalar @ARGV != 1 ) {
-        BarnOwl::error(
-            "Usage: jwrite JID [-t thread] [-s 'subject'] [-a account]");
-        return;
+        die("Usage: jwrite <jid> [-t <thread>] [-s <subject>] [-a <account>]\n");
     }
     else {
       $to = shift @ARGV;
@@ -497,16 +568,16 @@ sub cmd_jwrite {
     my @candidates = guess_jwrite($from, $to);
 
     unless(scalar @candidates) {
-        die("Unable to resolve JID $to");
+        die("Unable to resolve JID $to\n");
     }
 
     @candidates = grep {defined $_->[0]} @candidates;
 
     unless(scalar @candidates) {
         if(!$from) {
-            die("You must specify an account with -a");
+            die("You must specify an account with -a\n");
         } else {
-            die("Unable to resolve account $from");
+            die("Unable to resolve account $from\n");
         }
     }
 
@@ -522,6 +593,11 @@ sub cmd_jwrite {
         type    => $jwrite_type
     };
 
+    if (defined($jwrite_body)) {
+        process_owl_jwrite($jwrite_body);
+        return;
+    }
+
     if(scalar @candidates > 1) {
         BarnOwl::message(
             "Warning: Guessing account and/or destination JID"
@@ -532,11 +608,11 @@ sub cmd_jwrite {
            );
     }
 
-    my $cmd = "jwrite $jwrite_to -a $jwrite_from";
-    $cmd .= " -t $jwrite_thread" if $jwrite_thread;
-    $cmd .= " -s $jwrite_subject" if $jwrite_subject;
+    my @cmd = ('jwrite', $jwrite_to, '-a', $jwrite_from);
+    push @cmd, '-t', $jwrite_thread if $jwrite_thread;
+    push @cmd, '-s', $jwrite_subject if $jwrite_subject;
 
-    BarnOwl::start_edit_win($cmd, \&process_owl_jwrite );
+    BarnOwl::start_edit_win(BarnOwl::quote(@cmd), \&process_owl_jwrite);
 }
 
 sub cmd_jmuc {
@@ -558,8 +634,7 @@ sub cmd_jmuc {
     );
     my $func = $jmuc_commands{$cmd};
     if ( !$func ) {
-        BarnOwl::error("jmuc: Unknown command: $cmd");
-        return;
+        die("jmuc: Unknown command: $cmd\n");
     }
 
     {
@@ -581,7 +656,7 @@ sub cmd_jmuc {
             return unless $jid;
         }
         else {
-            BarnOwl::error('You must specify an account with -a {jid}');
+            die("You must specify an account with -a <jid>\n");
         }
         return $func->( $jid, $muc, @ARGV );
     }
@@ -594,7 +669,7 @@ sub jmuc_join {
     GetOptions( 'password=s' => \$password );
 
     $muc = shift @ARGV
-      or die("Usage: jmuc join MUC [-p password] [-a account]");
+      or die("Usage: jmuc join <muc> [-p <password>] [-a <account>]\n");
 
     die("Error: Must specify a fully-qualified MUC name (e.g. barnowl\@conference.mit.edu)\n")
         unless $muc =~ /@/;
@@ -607,6 +682,7 @@ sub jmuc_join {
                                                History  => {
                                                    MaxChars => 0
                                                   });
+    $completion_jids{$muc->GetJID('base')} = 1;
     return;
 }
 
@@ -614,12 +690,12 @@ sub jmuc_part {
     my ( $jid, $muc, @args ) = @_;
 
     $muc = shift @args if scalar @args;
-    die("Usage: jmuc part MUC [-a account]") unless $muc;
+    die("Usage: jmuc part [<muc>] [-a <account>]\n") unless $muc;
 
     if($conn->getConnectionFromJID($jid)->MUCLeave(JID => $muc)) {
         queue_admin_msg("$jid has left $muc.");
     } else {
-        die("Error: Not joined to $muc");
+        die("Error: Not joined to $muc\n");
     }
 }
 
@@ -629,7 +705,7 @@ sub jmuc_invite {
     my $invite_jid = shift @args;
     $muc = shift @args if scalar @args;
 
-    die('Usage: jmuc invite JID [muc] [-a account]')
+    die("Usage: jmuc invite <jid> [<muc>] [-a <account>]\n")
       unless $muc && $invite_jid;
 
     my $message = Net::Jabber::Message->new();
@@ -644,7 +720,7 @@ sub jmuc_invite {
 sub jmuc_configure {
     my ( $jid, $muc, @args ) = @_;
     $muc = shift @args if scalar @args;
-    die("Usage: jmuc configure [muc]") unless $muc;
+    die("Usage: jmuc configure [<muc>]\n") unless $muc;
     my $iq = Net::Jabber::IQ->new();
     $iq->SetTo($muc);
     $iq->SetType('set');
@@ -661,6 +737,7 @@ sub jmuc_presence_single {
     my @jids = $m->Presence();
 
     my $presence = "JIDs present in " . $m->BaseJID;
+    $completion_jids{$m->BaseJID} = 1;
     if($m->Anonymous) {
         $presence .= " [anonymous MUC]";
     }
@@ -684,7 +761,7 @@ sub jmuc_presence {
     my ( $jid, $muc, @args ) = @_;
 
     $muc = shift @args if scalar @args;
-    die("Usage: jmuc presence MUC") unless $muc;
+    die("Usage: jmuc presence [<muc>]\n") unless $muc;
 
     if ($muc eq '-a') {
         my $str = "";
@@ -699,7 +776,7 @@ sub jmuc_presence {
     }
     else {
         my $m = $conn->getConnectionFromJID($jid)->FindMUC(jid => $muc);
-        die("No such muc: $muc") unless $m;
+        die("No such muc: $muc\n") unless $m;
         BarnOwl::popless_ztext(jmuc_presence_single($m));
     }
 }
@@ -726,8 +803,7 @@ sub cmd_jroster {
     );
     my $func = $jroster_commands{$cmd};
     if ( !$func ) {
-        BarnOwl::error("jroster: Unknown command: $cmd");
-        return;
+        die("jroster: Unknown command: $cmd\n");
     }
 
     {
@@ -750,12 +826,39 @@ sub cmd_jroster {
             return unless $jid;
         }
         else {
-            BarnOwl::error('You must specify an account with -a {jid}');
+            die("You must specify an account with -a <jid>\n");
         }
         return $func->( $jid, $name, \@groups, $purgeGroups,  @ARGV );
     }
 }
 
+sub cmd_jaway {
+    my $cmd = shift;
+    local @ARGV = @_;
+    my $getopt = Getopt::Long::Parser->new;
+    my ($jid, $show);
+    my $p = new Net::Jabber::Presence;
+
+    $getopt->configure('pass_through', 'no_getopt_compat');
+    $getopt->getoptions(
+        'account=s' => \$jid,
+        'show=s'    => \$show
+    );
+    $jid ||= defaultJID();
+    if ($jid) {
+        $jid = resolveConnectedJID($jid);
+        return unless $jid;
+    }
+    else {
+        die("You must specify an account with -a <jid>\n");
+    }
+
+    $p->SetShow($show eq "online" ? "" : $show) if $show;
+    $p->SetStatus(join(' ', @ARGV)) if @ARGV;
+    $conn->getConnectionFromJID($jid)->Send($p);
+}
+
+
 sub jroster_sub {
     my $jid = shift;
     my $name = shift;
@@ -808,6 +911,9 @@ sub jroster_add {
     # Adding lots of users with the same name is a bad idea.
     $name = "" unless (1 == scalar(@ARGV));
 
+    $completion_jids{$baseJID} = 1;
+    $completion_jids{$name} = 1 if $name;
+
     foreach my $to (@ARGV) {
         my %jq  = $roster->query($to);
         my $iq = new Net::Jabber::IQ;
@@ -932,8 +1038,10 @@ sub process_owl_jwrite {
 
 sub process_incoming_chat_message {
     my ( $sid, $j ) = @_;
-    BarnOwl::queue_message( j2o( $j, { direction => 'in',
-                                   sid => $sid } ) );
+    if ($j->DefinedBody() || BarnOwl::getvar('jabber:spew') eq 'on') {
+        BarnOwl::queue_message( j2o( $j, { direction => 'in',
+                                           sid => $sid } ) );
+    }
 }
 
 sub process_incoming_error_message {
@@ -976,7 +1084,7 @@ sub process_incoming_normal_message {
     #	}
     #    }
     #
-    if(BarnOwl::getvar('jabber:spew') eq 'on') {
+    if ($j->DefinedBody() || BarnOwl::getvar('jabber:spew') eq 'on') {
         BarnOwl::queue_message( BarnOwl::Message->new(%jhash) );
     }
 }
@@ -989,12 +1097,14 @@ sub process_muc_presence {
 
 sub process_presence_available {
     my ( $sid, $p ) = @_;
-    my $from = $p->GetFrom();
+    my $from = $p->GetFrom('jid')->GetJID('base');
+    $completion_jids{$from} = 1;
+    return unless (BarnOwl::getvar('jabber:show_logins') eq 'on');
     my $to = $p->GetTo();
     my $type = $p->GetType();
     my %props = (
         to => $to,
-        from => $from,
+        from => $p->GetFrom(),
         recipient => $to,
         sender => $from,
         type => 'jabber',
@@ -1012,10 +1122,7 @@ sub process_presence_available {
         $props{body} = "$from is now offline. ";
         $props{loginout} = 'logout';
     }
-    $props{replysendercmd} = $props{replycmd} = "jwrite $from -i $sid";
-    if(BarnOwl::getvar('debug') eq 'on') {
-        BarnOwl::queue_message(BarnOwl::Message->new(%props));
-    }
+    BarnOwl::queue_message(BarnOwl::Message->new(%props));
 }
 
 sub process_presence_subscribe {
@@ -1032,8 +1139,8 @@ sub process_presence_subscribe {
 
     $props{body} = "Allow user ($from) to subscribe to your ($to) presence?\n" .
                    "(Answer with the `yes' or `no' commands)";
-    $props{yescommand} = "jroster auth $from -a $to";
-    $props{nocommand} = "jroster deauth $from -a $to";
+    $props{yescommand} = BarnOwl::quote('jroster', 'auth', $from, '-a', $to);
+    $props{nocommand} = BarnOwl::quote('jroster', 'deauth', $from, '-a', $to);
     $props{question} = "true";
     BarnOwl::queue_message(BarnOwl::Message->new(%props));
 }
@@ -1068,7 +1175,7 @@ sub process_presence_unsubscribe {
 
 sub process_presence_subscribed {
     my ( $sid, $p ) = @_;
-    queue_admin_msg("ignoring:".$p->GetXML());
+    queue_admin_msg("ignoring:".$p->GetXML()) if BarnOwl::getvar('jabber:spew') eq 'on';
     # RFC 3921 says we should respond to this with a "subscribe"
     # but this causes a flood of sub/sub'd presence packets with
     # some servers, so we won't. We may want to detect this condition
@@ -1078,7 +1185,7 @@ sub process_presence_subscribed {
 
 sub process_presence_unsubscribed {
     my ( $sid, $p ) = @_;
-    queue_admin_msg("ignoring:".$p->GetXML());
+    queue_admin_msg("ignoring:".$p->GetXML()) if BarnOwl::getvar('jabber:spew') eq 'on';
     # RFC 3921 says we should respond to this with a "subscribe"
     # but this causes a flood of unsub/unsub'd presence packets with
     # some servers, so we won't. We may want to detect this condition
@@ -1098,15 +1205,11 @@ sub process_presence_error {
 
 sub j2hash {
     my $j   = shift;
-    my %initProps = %{ shift() };
+    my %props = (type => 'jabber',
+                 dir  => 'none',
+                 %{$_[0]});
 
-    my $dir = 'none';
-    my %props = ( type => 'jabber' );
-
-    foreach my $k (keys %initProps) {
-        $dir = $initProps{$k} if ($k eq 'direction');
-        $props{$k} = $initProps{$k};
-    }
+    my $dir = $props{direction};
 
     my $jtype = $props{jtype} = $j->GetType();
     my $from = $j->GetFrom('jid');
@@ -1119,16 +1222,47 @@ sub j2hash {
     $props{sender}     = $from->GetJID('base');
     $props{subject}    = $j->GetSubject() if ( $j->DefinedSubject() );
     $props{thread}     = $j->GetThread() if ( $j->DefinedThread() );
-    $props{body}       = $j->GetBody() if ( $j->DefinedBody() );
+    if ( $j->DefinedBody() ) {
+        $props{body}   = $j->GetBody();
+        $props{body}  =~ s/\xEF\xBB\xBF//g; # Strip stray Byte-Order-Marks.
+    }
     $props{error}      = $j->GetError() if ( $j->DefinedError() );
     $props{error_code} = $j->GetErrorCode() if ( $j->DefinedErrorCode() );
     $props{xml}        = $j->GetXML();
 
-    if ( $jtype eq 'chat' ) {
-        $props{replycmd} =
-          "jwrite " . ( ( $dir eq 'in' ) ? $props{from} : $props{to} );
-        $props{replycmd} .=
-          " -a " . ( ( $dir eq 'out' ) ? $props{from} : $props{to} );
+    if ( $jtype eq 'groupchat' ) {
+        my $nick = $props{nick} = $from->GetResource();
+        my $room = $props{room} = $from->GetJID('base');
+        $completion_jids{$room} = 1;
+
+        my $muc;
+        if ($dir eq 'in') {
+            my $connection = $conn->getConnectionFromSid($props{sid});
+            $muc = $connection->FindMUC(jid => $from);
+        } else {
+            my $connection = $conn->getConnectionFromJID($props{from});
+            $muc = $connection->FindMUC(jid => $to);
+        }
+        $props{from} = $muc->GetFullJID($from) || $props{from};
+        $props{sender} = $nick || $room;
+        $props{recipient} = $room;
+
+        if ( $props{subject} && !$props{body} ) {
+            $props{body} =
+              '[' . $nick . " has set the topic to: " . $props{subject} . "]";
+        }
+    }
+    elsif ( $jtype eq 'headline' ) {
+        ;
+    }
+    elsif ( $jtype eq 'error' ) {
+        $props{body}     = "Error "
+          . $props{error_code}
+          . " sending to "
+          . $props{from} . "\n"
+          . $props{error};
+    }
+    else { # chat, or normal (default)
         $props{private} = 1;
 
         my $connection;
@@ -1147,49 +1281,16 @@ sub j2hash {
             $props{recipient} = $props{to}
               if ($connection->FindMUC(jid => $to));
         }
-    }
-    elsif ( $jtype eq 'groupchat' ) {
-        my $nick = $props{nick} = $from->GetResource();
-        my $room = $props{room} = $from->GetJID('base');
-        $props{replycmd} = "jwrite $room";
-        $props{replycmd} .=
-          " -a " . ( ( $dir eq 'out' ) ? $props{from} : $props{to} );
-	if ($props{subject}) {
-	  $props{replycmd} .= " -s " . $props{subject}
-	}
 
-        if ($dir eq 'out') {
-            $props{replysendercmd} = "jwrite ".$props{to}." -a ".$props{from};
+        # Populate completion.
+        if ($dir eq 'in') {
+            $completion_jids{ $props{sender} }= 1;
         }
         else {
-            $props{replysendercmd} = "jwrite ".$props{from}." -a ".$props{to};
-        }
-
-        $props{sender} = $nick || $room;
-        $props{recipient} = $room;
-
-        if ( $props{subject} && !$props{body} ) {
-            $props{body} =
-              '[' . $nick . " has set the topic to: " . $props{subject} . "]";
+            $completion_jids{ $props{recipient} } = 1;
         }
     }
-    elsif ( $jtype eq 'normal' ) {
-        $props{replycmd}  = "";
-        $props{private} = 1;
-    }
-    elsif ( $jtype eq 'headline' ) {
-        $props{replycmd} = "";
-    }
-    elsif ( $jtype eq 'error' ) {
-        $props{replycmd} = "";
-        $props{body}     = "Error "
-          . $props{error_code}
-          . " sending to "
-          . $props{from} . "\n"
-          . $props{error};
-    }
 
-    $props{replysendercmd} = $props{replycmd} unless $props{replysendercmd};
     return %props;
 }
 
@@ -1211,7 +1312,7 @@ sub getServerFromJID {
     if ($packet)    # Got srv record.
     {
         my @answer = $packet->answer;
-        return $answer[0]{target}, $answer[0]{port};
+        return $answer[0]->target, $answer[0]->port if @answer;
     }
 
     return $jid->GetServer(), 5222;
@@ -1231,6 +1332,7 @@ sub baseJID {
 
 sub resolveConnectedJID {
     my $givenJIDStr = shift;
+    my $loose = shift || 0;
     my $givenJID    = new Net::Jabber::JID;
     $givenJID->SetJID($givenJIDStr);
 
@@ -1238,40 +1340,63 @@ sub resolveConnectedJID {
     if ( $givenJID->GetResource() ) {
         # Specified account exists
         return $givenJIDStr if ($conn->jidExists($givenJIDStr) );
-        die("Invalid account: $givenJIDStr");
+        return resolveConnectedJID($givenJID->GetJID('base')) if $loose;
+        die("Invalid account: $givenJIDStr\n");
     }
 
     # Disambiguate.
     else {
-        my $matchingJID = "";
-        my $errStr =
-          "Ambiguous account reference. Please specify a resource.\n";
-        my $ambiguous = 0;
+        my $JIDMatchingJID = "";
+        my $strMatchingJID = "";
+        my $JIDMatches = "";
+        my $strMatches = "";
+        my $JIDAmbiguous = 0;
+        my $strAmbiguous = 0;
 
         foreach my $jid ( $conn->getJIDs() ) {
             my $cJID = new Net::Jabber::JID;
             $cJID->SetJID($jid);
             if ( $givenJIDStr eq $cJID->GetJID('base') ) {
-                $ambiguous = 1 if ( $matchingJID ne "" );
-                $matchingJID = $jid;
-                $errStr .= "\t$jid\n";
+                $JIDAmbiguous = 1 if ( $JIDMatchingJID ne "" );
+                $JIDMatchingJID = $jid;
+                $JIDMatches .= "\t$jid\n";
+            }
+            if ( $cJID->GetJID('base') =~ /$givenJIDStr/ ) {
+                $strAmbiguous = 1 if ( $strMatchingJID ne "" );
+                $strMatchingJID = $jid;
+                $strMatches .= "\t$jid\n";
             }
         }
 
         # Need further disambiguation.
-        if ($ambiguous) {
-            die($errStr);
-        }
-
-        # Not one of ours.
-        elsif ( $matchingJID eq "" ) {
-            die("Invalid account: $givenJIDStr");
+        if ($JIDAmbiguous) {
+            my $errStr =
+                "Ambiguous account reference. Please specify a resource.\n";
+            die($errStr.$JIDMatches);
         }
 
         # It's this one.
-        else {
-            return $matchingJID;
+        elsif ($JIDMatchingJID ne "") {
+            return $JIDMatchingJID;
         }
+
+        # Further resolution by substring.
+        elsif ($strAmbiguous) {
+            my $errStr =
+                "Ambiguous account reference. Please be more specific.\n";
+            die($errStr.$strMatches);
+        }
+
+        # It's this one, by substring.
+        elsif ($strMatchingJID ne "") {
+            return $strMatchingJID;
+        }
+
+        # Not one of ours.
+        else {
+            die("Invalid account: $givenJIDStr\n");
+        }
+
     }
     return "";
 }
@@ -1314,8 +1439,8 @@ sub guess_jwrite {
     my ($from_jid, $to_jid);
     my @matches;
     if($from) {
-        $from_jid = resolveConnectedJID($from);
-        die("Unable to resolve account $from") unless $from_jid;
+        $from_jid = resolveConnectedJID($from, 1);
+        die("Unable to resolve account $from\n") unless $from_jid;
         $to_jid = resolveDestJID($to, $from_jid);
         push @matches, [$from_jid, $to_jid] if $to_jid;
     } else {
@@ -1339,4 +1464,33 @@ sub guess_jwrite {
     return @matches;
 }
 
+################################################################################
+### Completion
+
+sub complete_user_or_muc { return keys %completion_jids; }
+sub complete_account { return $conn->getJIDs(); }
+
+sub complete_jwrite {
+    my $ctx = shift;
+    return complete_flags($ctx,
+                          [qw(-t -i -s)],
+                          {
+                              "-a" => \&complete_account,
+                          },
+                          \&complete_user_or_muc
+        );
+}
+
+sub complete_jabberlogout {
+    my $ctx = shift;
+    if($ctx->word == 1) {
+        return ("-A", complete_account() );
+    } else {
+        return ();
+    }
+}
+
+BarnOwl::Completion::register_completer(jwrite => sub { BarnOwl::Module::Jabber::complete_jwrite(@_) });
+BarnOwl::Completion::register_completer(jabberlogout => sub { BarnOwl::Module::Jabber::complete_jabberlogout(@_) });
+
 1;
